from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
import numpy.testing as npt
import numpy as np
from pyHalo.Halos.HaloModels.ULDM import ULDMFieldHalo, ULDMSubhalo
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
from lenstronomy.LensModel.Profiles.cnfw import CNFW
from astropy.cosmology import FlatLambdaCDM
from lenstronomy.LensModel.Profiles.uldm import Uldm
import pytest


class TestULDMHalo(object):

    def setup_method(self):
        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = None
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)

        self.mass = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = 100.0
        self.kwargs_profile = {'evaluate_mc_at_zlens': False,
                          'log10_m_uldm': -21, 'uldm_plaw': 1 / 3, 'scale_nfw': True}
        self.subhalo = ULDMSubhalo(self.mass, x, y, r3d, self.zhalo,
                                   True, self.lens_cosmo,
                                   self.kwargs_profile, self.truncation_class, self.concentration_class,
                       unique_tag=np.random.rand())
        self.fieldhalo = ULDMFieldHalo(self.mass, x, y, r3d, self.zhalo,
                                       False, self.lens_cosmo,
                                       self.kwargs_profile, self.truncation_class, self.concentration_class,
                                       unique_tag=np.random.rand())

    def test_lenstronomy_ID(self):

        ID = self.fieldhalo.lenstronomy_ID
        npt.assert_string_equal(ID[0], 'CNFW')
        npt.assert_string_equal(ID[1], 'ULDM')

        ID = self.subhalo.lenstronomy_ID
        npt.assert_string_equal(ID[0], 'CNFW')
        npt.assert_string_equal(ID[1], 'ULDM')

    def test_redshift_eval(self):

        z_subhalo = self.subhalo.z_eval
        z_field = self.fieldhalo.z_eval
        npt.assert_equal(z_field, self.zhalo)
        # because the concentration is evaluated at infall, and z_infall > z
        npt.assert_equal(True, z_subhalo > z_field)

    def test_profile_normalization_subhalo(self):
        """
        Test that the mass enclosed within r200 of the composite profile is correct
        and check that the ULDM core density is correct.
        """

        [cnfw_kwargs, uldm_kwargs] = self.subhalo.lenstronomy_params[0]
        Rs_angle, _ = self.lens_cosmo.nfw_physical2angle(10**8, self.subhalo.c, self.subhalo.z)
        sigma_crit = self.lens_cosmo.sigmacrit
        r200 = self.subhalo.c * Rs_angle

        M_nfw = CNFW().mass_3d_lens(r200, cnfw_kwargs['Rs'], cnfw_kwargs['alpha_Rs']*sigma_crit, cnfw_kwargs['r_core'])
        M_uldm = Uldm().mass_3d_lens(r200, uldm_kwargs['kappa_0']*sigma_crit, uldm_kwargs['theta_c'])

        npt.assert_almost_equal((M_uldm+M_nfw)/self.subhalo.mass,1,decimal=2) # less than 1% error
        _,theta_c,kappa_0 = self.subhalo.profile_args
        rho0 = Uldm().density_lens(0,uldm_kwargs['kappa_0'],
                                    uldm_kwargs['theta_c'])
        rhos = CNFW().density_lens(0,cnfw_kwargs['Rs'],
                                 cnfw_kwargs['alpha_Rs'],
                                 cnfw_kwargs['r_core'])
        rho_goal = Uldm().density_lens(0,kappa_0,theta_c)
        npt.assert_array_less(np.array([1-(rho0+rhos)/rho_goal]),np.array([0.1])) # less than 3% error

    def test_profile_normalization_fieldhalo(self):
        """
        Test that the mass enclosed within r200 of the composite profile is correct
        and check that the ULDM core density is correct.
        """

        [cnfw_kwargs, uldm_kwargs] = self.fieldhalo.lenstronomy_params[0]
        Rs_angle, _ = self.lens_cosmo.nfw_physical2angle(self.mass, self.fieldhalo.c, self.fieldhalo.z)
        sigma_crit = self.lens_cosmo.sigmacrit
        r200 = self.fieldhalo.c * Rs_angle

        M_nfw = CNFW().mass_3d_lens(r200, cnfw_kwargs['Rs'], cnfw_kwargs['alpha_Rs'] * sigma_crit,
                                    cnfw_kwargs['r_core'])
        M_uldm = Uldm().mass_3d_lens(r200, uldm_kwargs['kappa_0'] * sigma_crit, uldm_kwargs['theta_c'])

        npt.assert_almost_equal((M_uldm + M_nfw) / self.fieldhalo.mass, 1, decimal=2)  # less than 1% error
        _, theta_c, kappa_0 = self.fieldhalo.profile_args
        rho0 = Uldm().density_lens(0, uldm_kwargs['kappa_0'],
                                   uldm_kwargs['theta_c'])
        rhos = CNFW().density_lens(0, cnfw_kwargs['Rs'],
                                   cnfw_kwargs['alpha_Rs'],
                                   cnfw_kwargs['r_core'])
        rho_goal = Uldm().density_lens(0, kappa_0, theta_c)
        npt.assert_array_less(np.array([1 - (rho0 + rhos) / rho_goal]), np.array([0.1]))  # less than 10% error


if __name__ == '__main__':
   pytest.main()



import numpy as np
from pyHalo.Halos.halo_base import Halo

class PTMass(Halo):
    """
    Class that defines a point mass object in the lens model
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_clss = concentration_class
        mdef = 'PT_MASS'
        super(PTMass, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['POINT_MASS']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_params'):

            factor = self._lens_cosmo.point_mass_factor_z(self.z)

            theta_E = factor * np.sqrt(self.mass)

            kwargs = [{'center_x': self.x, 'center_y': self.y, 'theta_E': theta_E}]

            self._lenstronomy_params = kwargs

        return self._lenstronomy_params, None


    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ()



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.cnfw import CNFW
from lenstronomy.LensModel.Profiles.uldm import Uldm
from scipy.optimize import minimize
import lenstronomy.Util.constants as const
import numpy as np

class ULDMFieldHalo(Halo):

    """
    The base class for a truncated NFW halo
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        mdef = 'ULDM'
        super(ULDMFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['CNFW', 'ULDM']

    @property
    def c(self):
        """
        Computes the NFW halo concentration (once)
        """
        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        [concentration, theta_c, kappa_0] = self.profile_args

        x, y = np.round(self.x, 4), np.round(self.y, 4)
        Rs_angle, alpha_Rs = self._lens_cosmo.nfw_physical2angle(self.mass, concentration, self.z)
        kwargs_cnfw_temporary = {'alpha_Rs': alpha_Rs, 'Rs': Rs_angle,
            'center_x': x, 'center_y': y, 'r_core': 3*theta_c}
        kwargs_uldm_temporary = {'theta_c': theta_c, 'kappa_0': kappa_0,
            'center_x': x, 'center_y': y}
        kwargs = self._rescaled_cnfw_params(kwargs_cnfw_temporary,
                                                kwargs_uldm_temporary)
        return kwargs, None

    @property
    def z_eval(self):
        """
        Returns the halo redshift
        """
        return self.z

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            [theta_c, kappa_0] = self._uldm_args(self._args['log10_m_uldm'], self.mass, self._args['uldm_plaw'])
            self._profile_args = (self.c, theta_c, kappa_0)

        return self._profile_args

    def _uldm_args(self, m_log10, M, plaw):
        """
        :param m_log10: ULDM particle mass in log10 units
        :param M: ULDM virial mass
        :param plaw: power law exponent for core radius - virial mass relationship

        :return: core radius 'theta_c' and core density 'kappa_0' in lensing units.
        """
        r_c = self._core_radius(m_log10, M, plaw) # in kpc
        rho_c = self._central_density(m_log10, r_c) # in M_solar/pc^3
        r_c *= 10**3 # in pc
        Sigma_crit = self._lens_cosmo.sigma_crit_lensing * 10**(-12) # in M_solar/parsec^2
        D_lens = self._lens_cosmo.D_d * 10**6 # in pc
        theta_c = r_c / D_lens / const.arcsec # in arcsec
        kappa_0 = 429 * np.pi * rho_c * r_c / (2048 * np.sqrt(0.091) * Sigma_crit) # lensing units
        return [theta_c, kappa_0]

    def _core_radius(self, m_log10, M, plaw):
        """
        :param m_log10: ULDM particle mass in log10 units
        :param M: ULDM virial mass
        :param plaw: power law exponent for core radius - virial mass relationship

        :return: core radius in kpc, numerator of equation (3) in Schive et al. 2014 [1407.7762v2]
        """
        m22 = 10**(m_log10 + 22)
        a = 1/(1+self.z)
        M9 = 10**(np.log10(M) - 9)
        Zeta = (self._zeta(self.z) / self._zeta(0))
        return 1.6 * m22**(-1) * a**(1/2) * Zeta**(-1/6) * M9**(-plaw)

    def _central_density(self, m_log10, r_c):
        """
        :param m_log10: ULDM particle mass in log10 units
        :param r_c: core radius in kpc

        :return: central density in M_solar/pc^3, equation (7) in Schive et al. 2014 [1407.7762v2]
        """
        m23 = 10**(m_log10 + 23)
        a = 1/(1+self.z)
        x_c = r_c/a
        return 1.9 * a**(-1) * m23**(-2) * x_c**(-4)

    def _zeta(self,z):
        """
        :param z: redshift at which to evalute function

        :return: zeta function at z, see definition of ULDM density
        """
        Om_z = self._lens_cosmo.cosmo.astropy.Om(z)
        return (18*np.pi**2 + 82*(Om_z-1) - 39*(Om_z-1)**2) / Om_z

    def _rescaled_cnfw_params(self, cnfw_params, uldm_params):
        """
        :param cnfw_params: cored NFW halo lensing params
        :param uldm_params: ULDM halo lensing params

        :return: rescaled cored NFW params to fill up the remainder of the mass budget such that
        the composite profile has the inputted virial mass.
        """

        r200 = self._c * cnfw_params['Rs']
        rho0 = Uldm().density_lens(0,uldm_params['kappa_0'],
                                    uldm_params['theta_c'])
        rhos = CNFW().density_lens(0,cnfw_params['Rs'],
                                 cnfw_params['alpha_Rs'],
                                 cnfw_params['r_core'])

        args = (r200, self.mass, cnfw_params['Rs'], cnfw_params['alpha_Rs'],
                        uldm_params['kappa_0'], uldm_params['theta_c'],
                        rho0, rhos)
        initial_guess = np.array([0.9,1.1])
        bounds = ((0.5, 10), (0.5, 1.5))
        method = 'Nelder-Mead'
        out = minimize(self._function_to_minimize, initial_guess,
                           args, method=method, bounds=bounds, tol=0.001)
        beta, q = out['x']

        if beta<0:
            raise ValueError('Negative CNFW core radius, tweak your parameters.')
        elif q<0:
            raise ValueError('Negative ULDM profile mass, tweak your parameters.')
        else:
            pass

        cnfw_params['r_core'] /= beta
        uldm_params['kappa_0'] /= q

        M_nfw = CNFW().mass_3d_lens(r200, cnfw_params['Rs'],
                    cnfw_params['alpha_Rs']*self._lens_cosmo.sigmacrit, cnfw_params['r_core'])
        M_uldm = Uldm().mass_3d_lens(r200,
                    uldm_params['kappa_0']*self._lens_cosmo.sigmacrit, uldm_params['theta_c'])

        if (self._args['scale_nfw']):
            # When scale_nfw is True rescale alpha_Rs to improve mass accuracy
            scale = self.mass / (M_nfw+M_uldm)
            cnfw_params['alpha_Rs'] *= scale

        return [cnfw_params, uldm_params]

    def _constraint_mass(self, beta, q, r, m_target, rs, alpha_rs, kappa_0, theta_c):
        """
        :param beta: CNFW core radius ('r_core') rescaling parameter
        :param q: ULDM core density ('kappa_0') rescaling parameter
        :param r: r200 of CNFW profile
        :param m_target: halo virial mass
        :param rs: CNFW scale radius
        :param alpha_rs: CNFW deflection angle at rs, in absence of core
        :param kappa_0: ULDM core density
        :param theta_c: ULDM core radius

        :return: Evaluated mass constraint equation for CNFW component profile
        """
        r_core = beta * rs
        sigma_crit = self.lens_cosmo.sigmacrit
        args_nfw = (r, rs, alpha_rs*sigma_crit, r_core)
        args_uldm = (r, kappa_0*sigma_crit, theta_c)

        m_nfw = CNFW().mass_3d_lens(*args_nfw) / m_target
        m_uldm = q * Uldm().mass_3d_lens(*args_uldm) / m_target

        penalty = np.absolute(m_nfw + m_uldm - 1)
        if np.isnan(penalty):
            return 1e+12

        # penalize if not equal to zero
        return penalty

    def _constraint_density(self, beta, q, rho_target, rhos):
        """
        :param beta: CNFW core radius ('r_core') rescaling parameter
        :param q: ULDM core density ('kappa_0') rescaling parameter
        :param rho_target: ULDM density at r=0
        :param rhos: CNFW density at r=0

        :return: Evaluated density constraint equation for CNFW component profile
        """

        # penalize if not equal to zero
        return np.absolute(rhos - beta * rho_target * (q - 1))

    def _function_to_minimize(self, beta_q_args, r, m_target, rs, alpha_rs, kappa_0, theta_c, rho0, rhos):
        """
        :param beta_q_args: array containing beta, q parameters, see _constraint_mass and _constraint_density
        :param r: r200 of CNFW profile
        :param m_target: halo virial mass
        :param rs: CNFW scale radius
        :param alpha_rs: CNFW deflection angle at rs, in absence of core
        :param kappa_0: ULDM core density
        :param theta_c: ULDM core radius
        :param rho0: ULDM density at r=0
        :param rhos: CNFW density at r=0

        :return: Addition of mass and density constraints for CNFW component profile
        """

        # minimize will work with an array of arguments, so need to pass in an array and unpack it
        (beta, q) = beta_q_args

        constraint1 = self._constraint_mass(beta, q, r, m_target, rs, alpha_rs, kappa_0, theta_c)
        constraint2 = self._constraint_density(beta, q, rho0, rhos)

        return constraint1 + 20*constraint2

class ULDMSubhalo(ULDMFieldHalo):
    """
    Defines a composite ULDM+NFW halo that is a subhalo of the host dark matter halo. The only difference
    between the profile classes is that the subhalo will have it's concentration evaluated at infall redshift.
    """
    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval



import numpy.testing as npt
import numpy as np
from pyHalo.Halos.HaloModels.PTMass import PTMass
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import pytest

class TestPTMass(object):

    def setup_method(self):
        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = None
        self.concentration_class = None
        kwargs_profile = {}
        self.ptmass = PTMass(10**9, 1.0, 1.0, None, self.zhalo, False, self.lens_cosmo, kwargs_profile, None, None, 1.0)

    def test_lenstronomy_ID(self):

        id = self.ptmass.lenstronomy_ID
        npt.assert_string_equal(id[0], 'POINT_MASS')

    def test_profile_args(self):

        profile_args = self.ptmass.profile_args
        npt.assert_almost_equal(len(profile_args), 0)

    def test_thetaE(self):

        m = 10**8
        x = 0.0
        y = 0.0
        r3d = None
        is_subhalo = False
        kwargs_profile = {}
        unique_tag = 0.1
        ptmass = PTMass(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                      self.truncation_class, self.concentration_class, unique_tag)
        kwargs, _ = ptmass.lenstronomy_params
        theta_E = kwargs[0]['theta_E']
        theta_E_theory = self.lens_cosmo.point_mass_factor_z(ptmass.z) * np.sqrt(10**8)
        npt.assert_almost_equal(theta_E_theory, theta_E)


if __name__ == '__main__':
    pytest.main()




import numpy as np
from pyHalo.defaults import lenscone_default
from scipy.interpolate import interp1d
from pyHalo.Cosmology.cosmology import Cosmology
from scipy.integrate import quad
from pyHalo.Halos.lens_cosmo import LensCosmo
from scipy.integrate import simpson as simps
from pyHalo.concentration_models import preset_concentration_models
from lenstronomy.LensModel.lens_model import LensModel

class ITSampling(object):

    def __init__(self, x, cdf):
        """
        This class performs inverse transform sampling of a distribution given samples from the distribution
        :param samples: samples from the distribution
        """
        cdf = cdf / float(np.max(cdf))
        self._cdf_inverse = interp1d(cdf, x)
        self._umin = cdf[0]
        self._umax = cdf[-1]

    @classmethod
    def from_samples(cls, samples):
        """
        samples from the target distribution
        :param samples:
        :return:
        """
        ran = (np.min(samples), np.max(samples))
        h, x = np.histogram(samples, range=ran, bins=150)
        x = x[0:-1] + (x[1] - x[0]) / 2
        cdf = np.cumsum(h)
        cdf = cdf / float(np.max(cdf))
        return ITSampling(x, cdf)

    def __call__(self, n_samples):
        """
        Generates samples from the distribution
        :param n_samples: number of samples to draw
        :return: the samples
        """
        u = np.random.uniform(self._umin, self._umax, int(n_samples))
        samples_out = self._cdf_inverse(u)
        if n_samples == 1:
            return float(samples_out)
        else:
            return np.squeeze(samples_out)

def inverse_transform_sampling(x, function, args, n_samples):
    """

    :param x: the domain of the function across which you want to obtain samples
    :param function: the function or probability density you want to sample from
    :param args: arguments passed to function after x
    :param n_samples: number of samples to draw
    :return: samples from the probability density described by function
    """
    y = function(x, *args)
    cdf = np.cumsum(y)
    return inverse_transform_sampling_from_cdf(x, cdf, n_samples)

def inverse_transform_sampling_from_cdf(x, cdf, n_samples):
    """

    :param x: the domain of the function across which you want to obtain samples
    :param cdf: the cumulative distribution function of the pdf
    :param n_samples: number of samples to draw
    :return: samples from the probability density that corresponds to cdf
    """
    cdf = cdf * float(np.max(cdf)) ** -1
    cdf_inverse = interp1d(cdf, x)
    u = np.random.uniform(cdf[0], cdf[-1], n_samples)
    return cdf_inverse(u)

def generate_lens_plane_redshifts(zlens, zsource):
    """
    This routine sets up the redshift planes along the line of sight in the lens system
    :param zlens: main deflector plane redshift (if there is no main lens plane, then this can be set as None)
    :param zsource: source plane redshift
    :return: lens plane redshifts and the thickness of each slice
    """
    zmin = lenscone_default.default_zstart
    zstep = lenscone_default.default_z_step
    if zlens is None:
        redshifts = np.arange(zmin, zsource, zstep)
    else:
        front_z = np.arange(zmin, zlens, zstep)
        back_z = np.arange(zlens, zsource, zstep)
        redshifts = np.append(front_z, back_z)
    delta_zs = []
    for i in range(0, len(redshifts) - 1):
        delta_zs.append(redshifts[i + 1] - redshifts[i])
    delta_zs.append(zsource - redshifts[-1])
    return list(np.round(redshifts, 2)), np.round(delta_zs, 2)

def interpolate_ray_paths(x_coordinates, y_coordinates, lens_model, kwargs_lens, zsource,
                          terminate_at_source=False, source_x=None, source_y=None, evaluate_at_mean=False,
                          cosmo=None):

    """
    :param x_coordinates: x coordinates to interpolate (arcsec) (list)
    :param y_coordinates: y coordinates to interpolate (arcsec) (list)
    Typically x_coordinates/y_coordinates would be four image positions, or the coordinate of the lens centroid
    :param lens_model: instance of LensModel (lenstronomy)
    :param kwargs_lens: keyword arguments for lens model
    :param zsource: source redshift
    :param terminate_at_source: fix the final angular coordinate to the source coordinate
    :param source_x: source x coordinate (arcsec)
    :param source_y: source y coordinate (arcsec)
    :param evaluate_at_mean: if True, returns two single interp1d instances (one for each of x/y) that return the
    average of each individual x/y coordinate evaluated at each lens plane. For example, if you pass in four images positions
    the output would be an interpolation of the average x/y coordinate along the path traversed by the light
    (This is useful for aligning realizations with a background source significantly offset from the lens centroid)

    :return: Instances of interp1d (scipy) that return the angular coordinate of a ray given a
    comoving distance
    """

    angle_x = []
    angle_y = []

    if cosmo is None:
        cosmo = Cosmology()

    for i, (xpos, ypos) in enumerate(zip(x_coordinates, y_coordinates)):

        theta_x = [xpos]
        theta_y = [ypos]

        ray_x, ray_y, d = compute_comoving_ray_path(xpos, ypos, lens_model, kwargs_lens, zsource,
                                                          terminate_at_source, source_x, source_y, cosmo=cosmo)

        for rx, ry, di in zip(ray_x[1:], ray_y[1:], d[1:]):
            theta_x.append(rx / di)
            theta_y.append(ry / di)

        distances = [0.] + list(d[1:])
        distances = np.array(distances)
        theta_x = np.array(theta_x)
        theta_y = np.array(theta_y)

        angle_x.append(interp1d(distances, theta_x))
        angle_y.append(interp1d(distances, theta_y))

    if evaluate_at_mean:

        zrange = np.linspace(0., zsource, 100)
        comoving_distance_calc = cosmo.D_C_transverse
        distances = [comoving_distance_calc(zi) for zi in zrange]

        angular_coordinates_x = []
        angular_coordinates_y = []
        for di in distances:
            x_coords = [ray_x(di) for ray_x in angle_x]
            y_coords = [ray_y(di) for ray_y in angle_y]
            x_center = np.mean(x_coords)
            y_center = np.mean(y_coords)
            angular_coordinates_x.append(x_center)
            angular_coordinates_y.append(y_center)

        angle_x = [interp1d(distances, angular_coordinates_x)]
        angle_y = [interp1d(distances, angular_coordinates_y)]

    return angle_x, angle_y

def compute_comoving_ray_path(x_coordinate, y_coordinate, lens_model, kwargs_lens, zsource,
                              terminate_at_source=False, source_x=None, source_y=None, cosmo=None):

        """
        :param x_coordinate: x coordinates to interpolate (arcsec) (float)
        :param y_coordinate: y coordinates to interpolate (arcsec) (float)
        Typically x_coordinates/y_coordinates would be four image positions, or the coordinate of the lens centroid
        :param lens_model: instance of LensModel (lenstronomy)
        :param kwargs_lens: keyword arguments for lens model
        :param zsource: source redshift
        :param terminate_at_source: fix the final angular coordinate to the source coordinate
        :param source_x: source x coordinate (arcsec)
        :param source_y: source y coordinate (arcsec)
        :return: Instance of interp1d (scipy) that returns the angular coordinate of a ray given a
        comoving distance
        """

        if cosmo is None:
            cosmo = Cosmology()

        redshift_list = lens_model.redshift_list + [zsource]
        zstep = lenscone_default.default_z_step
        finely_sampled_redshifts = np.linspace(zstep, zsource - zstep, 50)
        all_redshifts = np.unique(np.append(redshift_list, finely_sampled_redshifts))

        all_redshifts_sorted = all_redshifts[np.argsort(all_redshifts)]

        comoving_distance_calc = cosmo.D_C_transverse

        x_start, y_start = 0., 0.
        z_start = 0.

        x_list = [0.]
        y_list = [0.]
        distances = [0.]

        alpha_x_start, alpha_y_start = x_coordinate, y_coordinate
        for zi in all_redshifts_sorted:

            x_start, y_start, alpha_x_start, alpha_y_start = lens_model.lens_model.ray_shooting_partial(x_start, y_start,
                                                                                alpha_x_start, alpha_y_start,
                                                                                z_start, zi, kwargs_lens)
            d = float(comoving_distance_calc(zi))
            x_list.append(x_start)
            y_list.append(y_start)
            distances.append(d)
            z_start = zi

        if terminate_at_source:
            d_src = comoving_distance_calc(zsource)
            x_list[-1] = source_x * d_src
            y_list[-1] = source_y * d_src

        return np.array(x_list), np.array(y_list), np.array(distances)

def de_broglie_wavelength(log10_m_uldm,v):
    '''
    Returns de Broglie wavelength of the ultra-light axion in kpc.

    :param log10_m_uldm: log(axion mass) in eV
    :param v: velocity in km/s
    '''
    m_axion=10**log10_m_uldm
    return 1.2*(1e-22/m_axion)*(100/v)

def delta_kappa(z_lens, z_source, m, rein, de_Broglie_wavelength):
    '''
    Returns standard deviation of the density fluctuations in projection in convergence units

    :param z_lens,z_source: lens and source redshifts
    :param m: main deflector halo mass in M_solar
    :param rein: Einstein radius in kpc
    :param de_Broglie_wavelength: de Broglie wavelength of axion in kpc
    '''
    l = LensCosmo(z_lens,z_source)
    sigma_crit = l.get_sigma_crit_lensing(z_lens, z_source) * (1e-3) ** 2
    ds = delta_sigma(m, z_lens, rein, de_Broglie_wavelength)
    delta_kappa = ds / sigma_crit
    return delta_kappa

def delta_sigma(m, z, rein, de_Broglie_wavelength):
    """
    Returns the mean ULDM fluctuation ampltiude of the host dark matter halo in units M_sun/kpc^2
    :param m:
    :param z:
    :param rein:
    :param de_Broglie_wavelength:
    :return:
    """
    cosmo = Cosmology()
    l = LensCosmo(z, 2.0, cosmo)
    model, _ = preset_concentration_models('DIEMERJOYCE19')
    cmodel = model(cosmo.astropy, scatter=False)
    c = cmodel.nfw_concentration(m, z)
    rhos, rs, _ = l.NFW_params_physical(m, c, z)
    nfw_rho_squared = projected_density_squared(rein, rhos, rs, c)
    delta_sigma = (np.sqrt(np.pi) * nfw_rho_squared * de_Broglie_wavelength)**0.5
    return delta_sigma

def projected_density_squared(R_ein, rhos, rs, concentration):

    '''
    Returns the integral along the line of sight of the square of the NFW profile
    i.e. integral rho_nfw(z)^2 dz

    :param R_ein: Einstein radius in kpc
    :param rhos: scale radius density of main deflector halo
    :param rs: scale radius of main deflector halo
    :param concentration: concentration of main deflector halo
    '''

    r200 = concentration * rs
    zmax = np.sqrt(r200 ** 2 - R_ein**2)

    x = lambda z: np.sqrt(R_ein ** 2 + z ** 2)/rs
    nfw_density_square = lambda z: rhos**2 / (x(z) * (1+x(z))**2)**2

    return 2 * quad(nfw_density_square, 0, zmax)[0]

def nfw_velocity_dispersion(rhos, rs, c, x=1):
    """
    Computes the central velocity dispersion of an NFW profile integrating from r = x * rs
    :param rhos:
    :param rs:
    :param c:
    :return:
    """
    G = 4.3e-6  # kpc/solMass * (km/sec)**2
    prefactor = 4 * np.pi * G * rhos ** 2 * rs ** 2
    density_at_r = rhos / (x * (1 + x) ** 2)
    _integrand = lambda x: (np.log(1 + x) - x / (1 + x)) / (x * (1 + x) ** 2) / x ** 2
    x = np.linspace(x, c, 150)
    y = _integrand(x)
    return np.sqrt(prefactor * simps(y, x) / density_at_r)

class MinHaloMassULDM(object):

    def __init__(self, log10_m_uldm, astropy_instance, log_mlow):
        """
        This class implements the minimum halo mass for ultra-light dark matter, given the particle mass and a cosmology

        The call method returns the maximum of log_mlow, and the resulting minimum ULDM halo mass to ensure that one does
        not inadvertently generate halos down to extremely low masses for heavy particles
        :param log10_m_uldm: particle mass
        :param astropy_instance: an instannce of astropy
        :param log_mlow: minimum halo mass to render, regardless of ULDM computation
        """
        m22 = 10**(log10_m_uldm + 22)
        self._Mmin0 = 4.4e7 * m22**(-3/2)
        self._astropy_instance = astropy_instance
        self._log_mlow = log_mlow

    def __call__(self, z):
        m_min = self.m_min(z)
        log10_m_min = np.log10(m_min)
        return max(log10_m_min, self._log_mlow)

    def m_min(self, z):
        return self._a(z) ** (-3 / 4) * (self._zeta(z) / self._zeta(0)) ** (1 / 4) * self._Mmin0

    def _a(self, z):
        return (1 + z) ** -1

    def _Om(self, z):
        return self._astropy_instance.Om(z)

    def _zeta(self, z):
        return (18 * np.pi ** 2 + 82 * (self._Om(z) - 1) - 39 * (self._Om(z) - 1) ** 2) / self._Om(z)

def multiplane_convergence(realization, cone_opening_angle_arcsec=2.5, npix=100, lens_model_list_macro=None,
                           kwargs_lens_macro=None, redshift_list_macro=None):
    """
    This function computes the effective multiplane convergence from a realization of dark matter halos.
    If no macromodel is specified (through lens_model_list_macro, kwargs_lens_macro, redshift_list_macro) then
    a typical one is assumed.

    :param realization: an instance of Realization
    :param cone_opening_angle_arcsec: the window size in arcsec (each axis spans 2*window_size)
    :param npix: number of pixels per axis
    :param lens_model_list_macro: list of lens models for the main deflector
    :param kwargs_lens_macro: keyword arguments for the macro lens models
    :param redshift_list_macro: list of redshifts for the macro lens models
    :return: the effective multiplane convergence, the full lens model used to create it, the keyword arguments for it,
    the macromodel lens model, and the keyword arguments for it
    """

    lens_model_list_halos, redshfit_array_halos, kwargs_lens_halos, _ = realization.lensing_quantities()
    if lens_model_list_macro is None:
        lens_model_list_macro = ['EPL', 'SHEAR']
        kwargs_lens_macro = [{'theta_E': 1.0, 'center_x': 0.0, 'center_y': 0.0, 'e1': 0.1, 'e2': -0.1,
                              'gamma': 2.0}, {'gamma1': 0.05, 'gamma2': 0.02}]
        redshift_list_macro = [realization.lens_cosmo.z_lens] * len(lens_model_list_macro)

    lens_model_macro = LensModel(lens_model_list_macro)
    lens_model = LensModel(lens_model_list_macro + lens_model_list_halos,
                           z_source=realization.lens_cosmo.z_source, multi_plane=True,
                           lens_redshift_list=redshift_list_macro + list(redshfit_array_halos),
                           cosmo=realization.lens_cosmo.cosmo.astropy)

    _r = np.linspace(-cone_opening_angle_arcsec/2, cone_opening_angle_arcsec/2, npix)
    xx, yy = np.meshgrid(_r, _r)
    kappa_macro = lens_model_macro.kappa(xx.ravel(), yy.ravel(), kwargs_lens_macro)
    kappa = lens_model.kappa(xx.ravel(), yy.ravel(), kwargs_lens_macro + kwargs_lens_halos)
    delta_kappa = (kappa - kappa_macro).reshape(npix, npix)
    return delta_kappa, lens_model, kwargs_lens_macro + kwargs_lens_halos, lens_model_macro, kwargs_lens_macro


def mask_annular(center_x, center_y, x_grid, y_grid, r_min, r_max = None):
    """
    An annular mask of specified inner and outer radii

    :param center_x: x-coordinate of center position of circular mask
    :param center_y: y-coordinate of center position of circular mask
    :param x_grid: x-coordinate grid
    :param y_grid: y-coordinate grid
    :param r_min: inner radius of mask in units of grid coordinates
    :param r_max: outer radius of mask in units of grid coordinates
    :return: mask array of shape x_grid with =0 inside the radius and =1 outside
    """
    x_shift = x_grid - center_x
    y_shift = y_grid - center_y
    R = np.sqrt(x_shift*x_shift + y_shift*y_shift)
    mask = np.ones(x_shift.shape)
    if r_max == None:
        mask[R <= r_min] = 0
    else:
        mask[R <= r_min] = 0
        mask[R >= r_max] = 0

    return mask



import pytest
import numpy.testing as npt
from pyHalo.Halos.tnfw_halo_util import *
from scipy.integrate import quad

class TestUtil(object):

    def test_tnfw_mass_fraction(self):

        def _integrand(ri, rhos, rs, rt):
            return 4 * np.pi * ri ** 2 * tnfw_density_profile(ri, rhos, rs, rt)
        tau = 8.0
        c = 15.0
        mfrac = tnfw_mass_fraction(tau, c)
        rs = 0.3
        rt = tau * rs
        r200 = rs * c
        rhos = 10 ** 7.5
        m200 = 4 * np.pi * rs ** 3 * rhos * (np.log(1+c) - c/(1+c))
        m = quad(_integrand, 0, r200, args=(rhos, rs, rt))[0]
        npt.assert_almost_equal(m/m200, mfrac)

    def test_interp(self):

        interpolator = tau_mf_interpolation()
        log10_c = 1.9
        m_final_over_m = 0.5
        log10_mfrac = np.log10(m_final_over_m)
        x = (log10_c, log10_mfrac)
        log10tau = interpolator(x)
        tau = 10 ** log10tau
        mfrac = tnfw_mass_fraction(tau, 10 ** log10_c)
        npt.assert_almost_equal(mfrac, m_final_over_m, 3)

if __name__ == '__main__':
     pytest.main()



import numpy as np
from pyHalo.Halos.HaloModels.TNFW import TNFWSubhalo

class TNFWFromParams(TNFWSubhalo):
    """
    Creates a TNFW halo based on physical params.
    """

    KEY_RT = "r_trunc_kpc"
    KEY_RS = "rs"
    KEY_RHO_S = "rhos"
    KEY_RV = "rv"
    KEY_Z_INFALL = "z_infall"
    KEY_ID = "index"

    def __init__(self, mass, x_kpc, y_kpc, r3d, z,sub_flag,
                 lens_cosmo_instance, args, unique_tag=None):
        """
        Defines a TNFW subhalo with physical params r_trunc_kpc, rs, rhos passed in the args argument
        """

        self._lens_cosmo = lens_cosmo_instance
        self._kpc_per_arcsec_at_z = self._lens_cosmo.cosmo.kpc_proper_per_asec(z)
        x = x_kpc / self._kpc_per_arcsec_at_z
        y = y_kpc / self._kpc_per_arcsec_at_z
        keys_physical = (self.KEY_RV,self.KEY_RS,self.KEY_RHO_S,self.KEY_RV,self.KEY_RT)
        self._params_physical = {key:args[key] for key in keys_physical}
        self._c = self._params_physical[self.KEY_RV] / self._params_physical[self.KEY_RS]
        self.id = args.get(self.KEY_ID)
        self._z_infall = args.get(self.KEY_Z_INFALL)
        super(TNFWFromParams, self).__init__(mass,x,y,r3d,z,sub_flag,lens_cosmo_instance,args,None,None,unique_tag)

    def density_profile_3d(self, r, params_physical=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """

        _params = self._params_physical if params_physical is None else params_physical

        r_t = _params[self.KEY_RT]
        r_s = _params[self.KEY_RS]
        rho_s = _params[self.KEY_RHO_S]

        n = 1

        x = r / r_s
        tau = r_t / r_s

        density_nfw = (rho_s / ((x)*(1+x)**2))

        return density_nfw * (tau**2 / (tau**2 + x**2))**n


    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            truncation_radius_kpc = self.params_physical[self.KEY_RT]
            self._profile_args = (self.c, truncation_radius_kpc)
        return self._profile_args


    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        KPC_TO_MPC = 1E-3

        if not hasattr(self, '_kwargs_lenstronomy'):

            r_t = self.params_physical[self.KEY_RT]
            r_s = self.params_physical[self.KEY_RS]
            rho_s = self.params_physical[self.KEY_RHO_S]

            Rs_angle, theta_Rs = self.lens_cosmo.nfw_physical2angle_fromNFWparams(rho_s *1 / KPC_TO_MPC**3,r_s * KPC_TO_MPC,self.z)

            x, y = np.round(self.x, 4), np.round(self.y, 4)

            Rs_angle = np.round(Rs_angle, 10)
            theta_Rs = np.round(theta_Rs, 10)
            r_trunc_arcsec = r_t / self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                      'center_x': x, 'center_y': y, 'r_trunc': r_trunc_arcsec}]

            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None



import numpy as np
from scipy import optimize
import re
import os, sys

# Set infall redshift.
zInfall = 3.0

# Set halo and subhalo properties
zHost = 0.5
MHost = 1.0e13
cHost = 8.0

MSub = 1.0e8
cSubMean = 20.0
cSubScatter = 0.000001  # (in dex)
# Set a lower limit on the halo concentration
cSubMinimum = 2.0

# Set the number of realizations
Ntree = 10

# Set ramdom seed.
randomSeed = 369

# Set cosmoligical parameters
HubbleConstant = 67.36
OmegaMatter = 0.3153
OmegaDarkEnergy = 0.6847

# Astronomical and physical constants (in units of kg, m, s)
G = 6.673e-11
Mpc = 3.08567758135e22
Msun = 1.98892e30
Gyr = 3.15581498e16
kilo = 1.0e3

# Critical density at the present day (in units of Msun/Mpc**3)
rhoCritical0 = 3.0 * (HubbleConstant * kilo / Mpc) ** 2 / (8.0 * np.pi * G) / (Msun / Mpc ** 3)


# Critical density at z (in units of Msun/Mpc**3)
def rhoCriticalZ(z):
    return rhoCritical0 * (OmegaDarkEnergy + OmegaMatter * (1.0 + z) ** 3)


# Convert redshift to time in Gyr
def t_z(z):
    return 2.0 / 3.0 / np.sqrt(OmegaDarkEnergy) \
           * np.arctanh(1.0 / np.sqrt(1.0 + OmegaMatter / OmegaDarkEnergy * (1.0 + z) ** 3)) \
           / (HubbleConstant * kilo / Mpc) \
           / Gyr


# Find the host mass when the subhalo infall assuming the host density profile
# does not change with time. (Note that the vriail radius, thus the halo mass,
# is redshift-dependent since the critical density of the Universe change with
# redshifts)
#
# Mass definition conversion.
def Mass_Definition_Conversion(M0, R0, c0, delta, z):
    def gc(x):
        return np.log(1.0 + x) - x / (1.0 + x)

    def Mean_Enclosed_Density(r):
        rs = R0 / c0
        Menclosed = M0 * gc(r / rs) / gc(c0)

        return Menclosed / (4.0 / 3.0 * np.pi * r ** 3)

    def rootFun(r):
        return Mean_Enclosed_Density(r) - delta * rhoCriticalZ(z)

    RNew = optimize.brentq(rootFun, 0.001 * R0, 10.0 * R0)
    cNew = RNew / R0 * c0
    MNew = M0 * gc(cNew) / gc(c0)

    return RNew, MNew, cNew


densityContrast = 200.0
RHost = (MHost / (4.0 / 3.0 * np.pi * densityContrast * rhoCriticalZ(zHost))) ** (1.0 / 3.0)
RsHost = RHost / cHost

# Sample the subhalo concentration.
cSub = np.zeros(Ntree)
if (cSubScatter > 0.0):
    cSub = cSubMean * 10.0 ** (cSubScatter * np.random.normal(0.0, 1.0, Ntree))
else:
    cSub = cSubMean

cSub = np.where(cSub > cSubMinimum, cSub, cSubMinimum)

RSubInfall = (MSub / (4.0 / 3.0 * np.pi * densityContrast * rhoCriticalZ(zInfall))) ** (1.0 / 3.0)
RsSubInfall = RSubInfall / cSub

RHostInfall, MHostInfall, cHostInfall = Mass_Definition_Conversion(MHost, RHost, cHost, 200.0, zInfall)

# Infall and current time.
tInfall = t_z(zInfall)
tCurrent = t_z(zHost)

# Generate parameter file.
folder = "Output_zInfall_" + str(zInfall) + "/cHost_" + str(cHost) + "/cSubhaloMean_" + str(cSubMean)

# Creat the folder containing the parameter file and output data.
os.makedirs(folder, exist_ok=True)

file_input = open("parameter_base.xml")
file_output = open(folder + "/parameter.xml", 'w')

line = file_input.readline()
while line:
    line = re.sub(r'randomSeed', str(randomSeed), line)

    line = re.sub(r'hubble', str(HubbleConstant), line)
    line = re.sub(r'Omega_M', str(OmegaMatter), line)
    line = re.sub(r'Omega_L', str(OmegaDarkEnergy), line)

    line = re.sub(r'redshiftInfall', str(zInfall), line)
    line = re.sub(r'hostConcentration', str(cHost), line)
    line = re.sub(r'subhaloConcentration', str(cSubMean), line)
    line = re.sub(r'timeInfall', str(tInfall), line)
    line = re.sub(r'timeCurrent', str(tCurrent), line)

    file_output.write(line, )
    line = file_input.readline()

file_input.close()
file_output.close()

# Generate tree file.
file_input = open("tree_base.xml")
file_output = open(folder + "/tree.xml", 'w')

# Write the headers.
for i in range(2):
    line = file_input.readline()
    file_output.write(line, )

line = '<trees>\n'
file_output.write(line, )

for i in range(Ntree):
    line = file_input.readline()
    while line:
        line = re.sub(r'timeInfall', str(tInfall), line)
        line = re.sub(r'timeCurrent', str(tCurrent), line)
        line = re.sub(r'MHostInfall', str(MHostInfall), line)
        line = re.sub(r'MHostCurrent', str(MHost), line)
        line = re.sub(r'RsHost', str(RsHost), line)

        line = re.sub(r'MSub', str(MSub), line)
        line = re.sub(r'RsSubInfall', str(RsSubInfall[i]), line)

        file_output.write(line, )
        line = file_input.readline()

    file_input.seek(0)
    for i in range(2):
        line = file_input.readline()

line = '</trees>'
file_output.write(line, )

file_input.close()
file_output.close()




from pyHalo.Halos.concentration import ConcentrationLudlow, ConcentrationDiemerJoyce, \
    ConcentrationPeakHeight, ConcentrationWDMHyperbolic, ConcentrationWDMPolynomial, ConcentrationLudlowWDM
from copy import deepcopy
import numpy as np


def preset_concentration_models(model_name, kwargs_model=None):
    """
    This function loads the concentration models, and implements some specific models from the literature
    :param model_name: the name of the concentration model (a string)
    :param kwargs_model: keyword arguments for the model class
    :return: the concentration model
    """
    if kwargs_model is None:
        kwargs_model_return = {}
    else:
        kwargs_model_return = deepcopy(kwargs_model)

    if model_name == 'DIEMERJOYCE19':
        return ConcentrationDiemerJoyce, kwargs_model_return
    elif model_name == 'LUDLOW2016':
        return ConcentrationLudlow, kwargs_model_return
    elif model_name == 'PEAK_HEIGHT_POWERLAW':
        return ConcentrationPeakHeight, kwargs_model_return
    elif model_name == 'WDM_HYPERBOLIC':
        return ConcentrationWDMHyperbolic, kwargs_model_return
    elif model_name == 'WDM_POLYNOMIAL':
        return ConcentrationWDMPolynomial, kwargs_model_return
    elif model_name == 'BOSE2016':
        # https://ui.adsabs.harvard.edu/abs/2016MNRAS.455..318B/abstract
        kwargs_model_return['concentration_cdm_class'] = ConcentrationDiemerJoyce
        kwargs_model_return['kwargs_cdm'] = {}
        kwargs_model_return['c_scale'] = 60.0
        kwargs_model_return['c_power'] = -0.17
        kwargs_model_return['c_power_inner'] = 1.0
        kwargs_model_return['mc_suppression_redshift_evolution'] = False
        return ConcentrationWDMPolynomial, kwargs_model_return
    elif model_name == 'LAROCHE2022':
        # https://ui.adsabs.harvard.edu/abs/2022MNRAS.517.1867L/abstract
        # calibrated for ultra-light dark matter
        kwargs_model_return['concentration_cdm_class'] = ConcentrationDiemerJoyce
        kwargs_model_return['kwargs_cdm'] = {}
        kwargs_model_return['c_scale'] = 21.42
        kwargs_model_return['c_power'] = -0.42
        kwargs_model_return['c_power_inner'] = 1.62
        return ConcentrationWDMPolynomial, kwargs_model_return
    elif model_name == 'LUDLOW_WDM':
        return ConcentrationLudlowWDM, kwargs_model_return
    elif model_name == 'FROM_FORMATION_HISTORY':
        norm, slope = 0.75, 0.4
        a = norm * (0.7 / kwargs_model['dlogT_dlogk']) ** slope
        norm, slope, shift = 0.94, 0.7, 0.6
        b = norm * abs(np.log(1.76 + shift) / np.log(kwargs_model['dlogT_dlogk'] + shift)) ** slope
        kwargs_model_return['a'] = a
        kwargs_model_return['b'] = b
        del kwargs_model_return['dlogT_dlogk']
        return ConcentrationWDMHyperbolic, kwargs_model_return
    elif model_name == 'CUSTOM':
        if not hasattr(kwargs_model_return['custom_class'], 'nfw_concentration'):
            raise Exception('a custom concentration-mass relation class must have a method nfw_concentration that'
                            'takes as input (halo mass, redshift) and returns the concentration')
        custom_mc_relation_class = kwargs_model_return['custom_class']
        del kwargs_model_return['custom_class']
        return custom_mc_relation_class, kwargs_model_return
    else:
        raise Exception('model name '+str(model_name)+' not recognized')



import pytest
import numpy as np
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Cosmology.geometry import Geometry
import numpy.testing as npt
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen, ShethTormenTurnover, ShethTormenMixedWDM
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw, WDMPowerLaw, MixedWDMPowerLaw
from colossus.lss.mass_function import massFunction

class TestShethTormen(object):

    def setup_method(self):

        z = 0.5
        delta_z = 0.02
        opening_angle = 16.0
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        _ = cosmo.colossus
        geometry_class = Geometry(cosmo, zlens, zsource, opening_angle, 'DOUBLE_CONE')
        delta_power_law_index = -0.1
        kwargs_model = {}
        kwargs_model['m_pivot'] = 10.0**8
        kwargs_model['log_mlow'] = 6.0
        kwargs_model['log_mhigh'] = 10.0
        kwargs_model['LOS_normalization'] = 1.0
        kwargs_model['draw_poisson'] = False
        kwargs_model['delta_power_law_index'] = delta_power_law_index

        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 20)
        m_h = m * h

        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / kwargs_model['m_pivot']), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + delta_power_law_index
        norm_dv = 10 ** coeffs[1]
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = 1.0 * norm_dv * volume_element_comoving
        normalization *= 1/(kwargs_model['m_pivot']**plaw_index)

        draw_poisson = False
        self.mass_function = CDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization)
        draw_poisson = True
        self.mass_function_poisson = CDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization)

        self._kwargs_from_redshift = {'z': z, 'delta_z': delta_z, 'geometry_class': geometry_class,
                                      'kwargs_model': kwargs_model}

    def test_draw_poisson(self):

        m = self.mass_function.draw()
        m_poisson = self.mass_function_poisson.draw()
        n1 = len(m)
        n2 = len(m_poisson)
        npt.assert_equal(True, n1!=n2)

    def test_from_redshift(self):
        mfunc = ShethTormen.from_redshift(**self._kwargs_from_redshift)
        m = mfunc.draw()
        npt.assert_almost_equal(len(m) / len(self.mass_function.draw()), 1, 3)

class TestShethTormenTurnover(object):

    def setup_method(self):

        z = 0.5
        delta_z = 0.02
        opening_angle = 16.0
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        geometry_class = Geometry(cosmo, zlens, zsource, opening_angle, 'DOUBLE_CONE')
        colossus_cosmology = cosmo.colossus
        np.random.seed(100)
        kwargs_model = {}
        delta_power_law_index = 0.0
        kwargs_model['m_pivot'] = 10.0 ** 8
        kwargs_model['log_mlow'] = 6.0
        kwargs_model['log_mhigh'] = 10.0
        kwargs_model['draw_poisson'] = False
        kwargs_model['LOS_normalization'] = 1.0
        kwargs_model['delta_power_law_index'] = delta_power_law_index
        kwargs_model['log_mc'] = 7.0
        kwargs_model['a_wdm'] = 1.0
        kwargs_model['b_wdm'] = 1.0
        kwargs_model['c_wdm'] = -2.5
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 20)
        m_h = m * h
        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / kwargs_model['m_pivot']), np.log10(dndM_comoving), 1)

        plaw_index = coeffs[0] + delta_power_law_index
        norm_dv = 10 ** coeffs[1]
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = 1.0 * norm_dv * volume_element_comoving /(kwargs_model['m_pivot']**plaw_index)

        draw_poisson = False
        self.mass_function = WDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization, kwargs_model['log_mc'], kwargs_model['a_wdm'],
                                         kwargs_model['b_wdm'], kwargs_model['c_wdm'])
        draw_poisson = True
        self.mass_function_poisson = WDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization, kwargs_model['log_mc'], kwargs_model['a_wdm'],
                                         kwargs_model['b_wdm'], kwargs_model['c_wdm'])

        self._kwargs_from_redshift = {'z': z, 'delta_z': delta_z, 'geometry_class': geometry_class,
                                      'kwargs_model': kwargs_model}

    def test_draw_poisson(self):

        m = self.mass_function.draw()
        m_poisson = self.mass_function_poisson.draw()
        npt.assert_equal(True, len(m)!=len(m_poisson))

    def test_from_redshift(self):
        mfunc = ShethTormenTurnover.from_redshift(**self._kwargs_from_redshift)
        m = mfunc.draw()
        n1 = len(m)
        n2 = len(self.mass_function.draw())
        npt.assert_equal(True, n1==n2)

class TestShethTormenMixedDM(object):

    def setup_method(self):

        z = 0.5
        delta_z = 0.02
        opening_angle = 8.0
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        geometry_class = Geometry(cosmo, zlens, zsource, opening_angle, 'DOUBLE_CONE')
        _ = cosmo.colossus
        #np.random.seed(101)
        kwargs_model = {}
        delta_power_law_index = -0.1
        kwargs_model['m_pivot'] = 10.0 ** 8
        kwargs_model['log_mlow'] = 6.0
        kwargs_model['log_mhigh'] = 10.0
        kwargs_model['draw_poisson'] = False
        kwargs_model['LOS_normalization'] = 1.0
        kwargs_model['delta_power_law_index'] = delta_power_law_index
        kwargs_model['log_mc'] = 7.0
        kwargs_model['a_wdm'] = 1.0
        kwargs_model['b_wdm'] = 1.0
        kwargs_model['c_wdm'] = -1.2
        kwargs_model['mixed_DM_frac'] = 0.6
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 20)
        m_h = m * h
        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / kwargs_model['m_pivot']), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + delta_power_law_index
        norm_dv = 10 ** coeffs[1]
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = 1.0 * norm_dv * volume_element_comoving /(kwargs_model['m_pivot']**plaw_index)
        draw_poisson = False

        self.mass_function = MixedWDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization, kwargs_model['log_mc'], kwargs_model['a_wdm'],
                                         kwargs_model['b_wdm'], kwargs_model['c_wdm'], kwargs_model['mixed_DM_frac'])
        draw_poisson = True
        self.mass_function_poisson = MixedWDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, normalization, kwargs_model['log_mc'], kwargs_model['a_wdm'],
                                         kwargs_model['b_wdm'], kwargs_model['c_wdm'], kwargs_model['mixed_DM_frac'])

        self._kwargs_from_redshift = {'z': z, 'delta_z': delta_z, 'geometry_class': geometry_class,
                                      'kwargs_model': kwargs_model}

    def test_draw_poisson(self):

        m = self.mass_function.draw()
        m_poisson = self.mass_function_poisson.draw()
        n1 = len(m)
        n2 = len(m_poisson)
        npt.assert_equal(True, n1!=n2)

    def test_from_redshift(self):
        n_iter = 250
        n1 = 0
        n2 = 0
        for i in range(0, n_iter):
            mfunc = ShethTormenMixedWDM.from_redshift(**self._kwargs_from_redshift)
            m = mfunc.draw()
            n1 += len(m)
            n2 += len(self.mass_function.draw())
        n1 /= n_iter
        n2 /= n_iter
        npt.assert_almost_equal(n1/n2, 1, 2)

class TestSampling(object):

    def setup_method(self):

        z = 0.5
        delta_z = 0.02
        opening_angle = 16.0
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        _ = cosmo.colossus
        geometry_class = Geometry(cosmo, zlens, zsource, opening_angle, 'DOUBLE_CONE')
        kwargs_model = {}
        kwargs_model['m_pivot'] = 10.0**8
        kwargs_model['log_mlow'] = 6.0
        kwargs_model['log_mhigh'] = 10.0
        kwargs_model['draw_poisson'] = False
        delta_power_law_index = -0.1
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 20)
        m_h = m * h

        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / kwargs_model['m_pivot']), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + delta_power_law_index
        norm_dv = 10 ** coeffs[1]
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = 1.0 * norm_dv * volume_element_comoving
        normalization *= 1/(kwargs_model['m_pivot']**plaw_index)

        draw_poisson = False
        self.mass_function = CDMPowerLaw(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         draw_poisson, 10*normalization)
        self._plaw_index_theory = plaw_index


    def test_sampling(self):

        m = self.mass_function.draw()
        h, b = np.histogram(np.log10(m), range=(6, 8), bins=20)
        log10h = np.log10(h)
        log10m = b[0:-1]
        coefs = np.polyfit(log10m, log10h, 1)
        npt.assert_almost_equal((self._plaw_index_theory+1)/coefs[0], 1, 2)

if __name__ == '__main__':
   pytest.main()



import numpy.testing as npt
import numpy as np
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy
from pyHalo.Halos.HaloModels.PsuedoJaffe import PJaffeSubhalo, PJaffeFieldhalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import TruncationRoche
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import pytest
from lenstronomy.LensModel.Profiles.pseudo_jaffe import PseudoJaffe as PJaffeLenstronomy

class TestPjaffeHalo(object):

    def setup_method(self):
        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = TruncationRoche(None, 100000000.0)
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)
        self.lclenstronomy = LensCosmoLenstronomy(self.zhalo, self.zsource, astropy)

        m = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = 100.0
        is_subhalo = True
        kwargs_profile = {'evaluate_mc_at_zlens': False}
        unique_tag = 1.0
        self.subhalo = PJaffeSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                      self.truncation_class, self.concentration_class, unique_tag)
        is_subhalo = False
        unique_tag = 2.0
        self.field_halo = PJaffeFieldhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                     self.truncation_class, self.concentration_class, unique_tag)

    def test_lenstronomy_ID(self):
        id = self.subhalo.lenstronomy_ID
        npt.assert_string_equal(id[0], 'PJAFFE')
        id = self.field_halo.lenstronomy_ID
        npt.assert_string_equal(id[0], 'PJAFFE')

    def test_z_infall(self):

        z_infall = self.subhalo.z_infall
        npt.assert_equal(True, self.zhalo <= z_infall)

    def test_total_mass(self):

        c = float(self.subhalo.profile_args)
        rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(self.subhalo.mass, c, self.zhalo)
        fc = np.log(1 + c) - c / (1 + c)
        m_nfw = 4 * np.pi * rs ** 3 * rhos * fc

        lenstronomy_kwargs, _ = self.subhalo.lenstronomy_params
        sigma0, ra, rs = lenstronomy_kwargs[0]['sigma0'], lenstronomy_kwargs[0]['Ra'], lenstronomy_kwargs[0]['Rs']

        arcsec_to_kpc = self.lens_cosmo.cosmo.kpc_proper_per_asec(self.zhalo)
        ra *= arcsec_to_kpc ** -1
        rs *= arcsec_to_kpc ** -1
        rho = self.subhalo._rho(m_nfw, rs, ra, c*rs)

        pjaffe_lenstronomy = PJaffeLenstronomy()
        m3d = pjaffe_lenstronomy.mass_3d(c*rs, rho, ra, rs)
        npt.assert_almost_equal(np.log10(m3d), np.log10(m_nfw))

    def test_concentration(self):

        profile_args = self.subhalo.profile_args
        (c_subhalo) = profile_args
        profile_args = self.field_halo.profile_args
        (c_fieldhalo) = profile_args
        npt.assert_equal(c_subhalo/c_fieldhalo<1, True)

    def test_params_physical(self):

        params_physical = self.subhalo.params_physical
        npt.assert_equal(len(params_physical), 4)

if __name__ == '__main__':
    pytest.main()




from pyHalo.PresetModels.cdm import CDM
from pyHalo.realization_extensions import RealizationExtensions


def SIDM_core_collapse(z_lens, z_source, mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos, kwargs_sub_function=None,
        kwargs_field_function=None, sigma_sub=0.025, log10_sigma_sub=None, log_mlow=6., log_mhigh=10.,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None, collapsed_halo_profile='SPL_CORE',
        kwargs_collapsed_profile={'x_core_halo': 0.05, 'x_match': 3.0, 'log_slope_halo': 3.0}):

    """
    Generates realizations of SIDM given the fraction of core-collapsed halos as a function of halo mass

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param mass_ranges_subhalos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_subhalos = [[6, 8], [8, 9], ...]
    :param mass_ranges_field_halos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_field_halos = [[6, 8], [8, 9], ...]
    :param probabilities_subhalos: a list of functions, or a list of the fraction of collapsed subhalos in each mass bin
    :param probabilities_field_halos: a list of functions, or a list of the fraction of collapsed field halos in each
    mass bin
    :param kwargs_sub_function: if probabilities_subhalos is a list of functions, specifies keyword arguments
    for each function
    :param kwargs_field_function: if probabilities_field_halos is a list of functions, specifies keyword arguments
    for each function
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: optional setting of sigma_sub in log10-scale (useful for log-uniform priors); if this is specified
    it overwrites the value of sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight mass function
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param collapsed_halo_profile: string that sets the density profile of core-collapsed halos
    currently implemented models are SPL_CORE and GNFW (see example notebook)
    :param kwargs_collapsed_profile: keyword arguments for the collapsed profile (see example notebook)
    :return: a realization of dark matter structure in SIDM
    """

    two_halo_contribution = True
    delta_power_law_index = 0.0
    cdm = CDM(z_lens, z_source, sigma_sub, log_mlow, log_mhigh, log10_sigma_sub,
              concentration_model_subhalos, kwargs_concentration_model_subhalos,
              concentration_model_fieldhalos, kwargs_concentration_model_fieldhalos,
              truncation_model_subhalos, kwargs_truncation_model_subhalos,
              truncation_model_fieldhalos, kwargs_truncation_model_fieldhalos,
              shmf_log_slope, cone_opening_angle_arcsec, log_m_host, r_tidal,
              LOS_normalization, two_halo_contribution, delta_power_law_index,
              geometry_type, kwargs_cosmo)

    extension = RealizationExtensions(cdm)
    index_collapsed = extension.core_collapse_by_mass(mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos, kwargs_sub_function, kwargs_field_function)
    sidm = extension.add_core_collapsed_halos(index_collapsed, collapsed_halo_profile, **kwargs_collapsed_profile)
    return sidm



import h5py
import numpy as np
from typing import Iterable

class GalacticusUtil():
    """
    This class provides utility for working with galacticus files.
    """

    #Names of galacticus parameters
    #Position of the subhalo relative to the top level host halo
    PARAM_X = "positionOrbitalX"
    """The GALACTICUS output parameter for the X coordinate of the subhalo relative to the main halo"""
    PARAM_Y = "positionOrbitalY" 
    """The GALACTICUS output parameter for the Y coordinate of the subhalo relative to the main halo"""
    PARAM_Z = "positionOrbitalZ"
    """The GALACTICUS output parameter for the Z coordinate of the subhalo relative to the main halo"""

    #Position of the subhalo relative to it's host halo or subhalo (not top level)
    PARAM_X_REL = "satellitePositionX"

    PARAM_Y_REL = "satellitePositionY"

    PARAM_Z_REL = "satellitePositionZ"


    PARAM_MASS_BOUND = "satelliteBoundMass"

    PARAM_MASS_INFALL = "massHaloEnclosedCurrent"
    """The infall mass of the subhalo"""

    PARAM_ISOLATED = "nodeIsIsolated"

    PARAM_HIERARCHY = "nodeHierarchyLevel"

    PARAM_RADIUS_VIRIAL = 'darkMatterOnlyRadiusVirial'

    PARAM_SPHERE_RADIUS = "spheroidRadius"

    PARAM_SPHERE_ANGULAR_MOMENTUM = "spheroidAngularMomentum"

    PARAM_SPHERE_MASS_STELLAR = "spheroidMassStellar"

    PARAM_SPHERE_MASS_GAS = "spheroidMassGas"

    PARAM_RADIUS_SCALE = "darkMatterProfileScaleRadius"

    PARAM_DENSITY_PROFILE_RADIUS = "densityProfileRadius"

    PARAM_DENSITY_PROFILE = "densityProfile"

    PARAM_Z_LAST_ISOLATED = "redshiftLastIsolated"

    PARAM_TNFW_RADIUS_TRUNCATION = "radiusTidalTruncationNFW"
    PARAM_TNFW_RHO_S = "densityNormalizationTidalTruncationNFW"

    PARAM_TREE_ORDER = "custom_treeOutputOrder"
    PARAM_TREE_INDEX = "custom_treeIndex"
    PARAM_NODE_ID = "id"

    PARAM_CONCENTRATION = "concentration"

    HDF5_GROUP_OUTPUT_PRIMARY = "Outputs"
    """
    The name of the primary output group in a GALACTICUS output file. 
    This is the primary output group that contains nodedata at different output times
    """

    HDF5_GROUP_OUTPUT_N_PREFIX = "Output"
    """The name of the prefix of the "OutputN" groups in GALACTICUS output file"""

    HDF5_GROUP_NODEDATA = "nodeData"
    """The name of the hdf5 group containing nodedata."""

    HDF5_DSET_TREECOUNT = "mergerTreeCount"

    HDF5_DSET_TREEINDEX = "mergerTreeIndex"

    HDF5_DSET_TREESTART = "mergerTreeStartIndex"

    def hdf5_read_output_indices(self,f):
        """
        Returns the numbers assigned to the various galacticus outputs.
        For if the Galacticus file has the following groups: Outputs/Output1, Outputs/Output10
        this function will return [1,10]

        :param f: h5py.File read from Galacticus output
        """
        group_outputs = f[self.HDF5_GROUP_OUTPUT_PRIMARY]

        outputs = []

        trim = len(self.HDF5_GROUP_OUTPUT_N_PREFIX)
        for key in group_outputs.keys():
            try:
                if(key[:trim]) != self.HDF5_GROUP_OUTPUT_N_PREFIX:
                    continue
                outputs.append(int(key[trim:]))
            except(ValueError):
                pass
        
        return np.array(outputs)
    

    def hdf5_access_output_n_nodedata(self,f,output_n):
        """
        Returns the Outputs/OutputN/nodedata groups

        :param f: h5py.File read from Galacticus output
        :param output_n: The number corresponding to the output to be read
        """
        return f[self.HDF5_GROUP_OUTPUT_PRIMARY][f"{self.HDF5_GROUP_OUTPUT_N_PREFIX}{output_n}"][self.HDF5_GROUP_NODEDATA]


    def hdf5_read_dset(self,dset):
        """
        Reads a hdf5 dataset into a numpy array
        """
        arr = np.zeros(dset.shape,dtype=dset.dtype)
        dset.read_direct(arr)
        return arr

    def hdf5_read_custom_nodedata(self,f,output_index):
        """
        To make analysis more convenient, it is useful to add custom datasets to those present in the Outputs/OutputN/nodedata.
        This function is used when read_nodedata_galacticus is called to create custom nodedata datasets for convenience.

        :param f: h5py.File read from Galacticus output
        :param output_index: The output to read indices for.
        """
        group_outputn = f[self.HDF5_GROUP_OUTPUT_PRIMARY][f"{self.HDF5_GROUP_OUTPUT_N_PREFIX}{output_index}"]

        tree_index = self.hdf5_read_dset(group_outputn[self.HDF5_DSET_TREEINDEX])
        tree_start = self.hdf5_read_dset(group_outputn[self.HDF5_DSET_TREESTART])

        total_count = self.hdf5_read_nodecount_total(f,output_index)

        node_tree_index = np.zeros(total_count,dtype=int)
        node_tree_order = np.zeros(total_count,dtype=int)

        for n in range(1,len(tree_start)):
            start,stop = tree_start[n-1],tree_start[n]
            node_tree_order[start:stop] = n - 1
            node_tree_index[start:stop] = tree_index[n - 1] 

        node_tree_order[stop:] = n
        node_tree_index[stop:] = tree_index[n]

        node_index = np.arange(total_count)

        return {self.PARAM_TREE_INDEX:node_tree_index,
                self.PARAM_TREE_ORDER:node_tree_order,
                self.PARAM_NODE_ID:node_index}
    
    def hdf5_read_nodecount_total(self,f,output_index):
        """
        Returns the total number of nodes at a given output.

        :param f: h5py.File read from Galacticus output
        :param output_index: The output index to read from.
        """

        group_outputn = f[self.HDF5_GROUP_OUTPUT_PRIMARY][f"{self.HDF5_GROUP_OUTPUT_N_PREFIX}{output_index}"]

        tree_count = self.hdf5_read_dset(group_outputn[self.HDF5_DSET_TREECOUNT])

        return np.sum(tree_count)
    
    def read_nodedata_galacticus(self,path, output_index = None,
                                params_to_read = None,
                                nodes_only=True):
        """
        Reads a galacticus output file at a given path, returns nodedata for a specified tree.
        Galacticus: https://github.com/galacticusorg/galacticus

        Note: if your build of galacticus has a different output formats you can inherit the GalacticusUtil class and modify
        parameter names / methods used when reading galacticus output. 

        :param path: Path to nodedata
        :param output_index: The index of the tree to read. If None defaults to the final output.
        :param params_to_read: A list of parameters to read, if None reads all parameters.
        :nodes_only: If true only reads nodedata entries that have entry for each node, if false reads all datasets in the nodedata group.
        """
        with h5py.File(path,"r") as f:
            return self.hdf5_read_galacticus_nodedata(f,output_index,params_to_read,nodes_only)


    def hdf5_read_galacticus_nodedata(self,f, output_index = None,
                                params_to_read = None,
                                nodes_only=True):
        """
        Reads a galacticus output given a h5py.File read from galacticus output
        Galacticus: https://github.com/galacticusorg/galacticus

        Note: if your build of galacticus has a different output formats you can inherit the GalacticusUtil class and modify
        parameter names / methods used when reading galacticus output. 

        :param f: A h5py.File read from galacticus output.
        :param output_index: The index of the tree to read. If None defaults to the final output.
        :param params_to_read: A list of parameters to read, if None reads all parameters.
        :nodes_only: If true only reads nodedata entries that have entry for each node, if false reads all datasets in the nodedata group.
        """

        #If no output is specified read the final output
        if output_index is None:

            outputs_ns = self.hdf5_read_output_indices(f)

            #If no suitable outputs are found, return None

            if len(outputs_ns) == 0:

                return None
            
            output_index = np.max(outputs_ns)

        group_nodedata = self.hdf5_access_output_n_nodedata(f,output_index)

        nodedata = {}

        #Allow user to pass a single parameter to read as just a string

        if isinstance(params_to_read,str):

            params_to_read = [params_to_read] 

        nodecount = self.hdf5_read_nodecount_total(f,output_index)

        #Loop through datasets in nodedata

        for key in group_nodedata.keys():

            dset = group_nodedata[key]

            #Read dataset if it is a dataset and it is contained in the params_to_read variable or that variable is None
            if isinstance(dset,h5py.Dataset) and (params_to_read is None or key in params_to_read):
                if not nodes_only or (nodes_only and dset.shape[0] == nodecount):
                    nodedata[key] = self.hdf5_read_dset(dset)

        
        custom = self.hdf5_read_custom_nodedata(f,output_index)
        return nodedata | custom




"""
These functions define some preset halo mass function models that can be specified instead of individually specifying
a bunch of keyword arguments. This is intended for a quick and easy user interface. The definitions shown here can also
act as a how-to guide if one wants to explore more complicated descriptions of the halo mass function, as the models
presented here show what each keyword argument accepted by pyHalo does.
"""

__all__ = ['preset_model_from_name']

def preset_model_from_name(name):
    """
    Retruns a preset_model function from a string
    :param name: the name of the preset model, should be the name of a function in this file
    :return: the function
    """
    if name == 'CDM':
        from pyHalo.PresetModels.cdm import CDM
        return CDM
    elif name == 'WDM':
        from pyHalo.PresetModels.wdm import WDM
        return WDM
    elif name == 'SIDM_core_collapse':
        from pyHalo.PresetModels.sidm import SIDM_core_collapse
        return SIDM_core_collapse
    elif name == 'ULDM':
        from pyHalo.PresetModels.uldm import ULDM
        return ULDM
    elif name == 'CDMEmulator':
        from pyHalo.PresetModels.external import CDMFromEmulator
        return CDMFromEmulator
    elif name == 'WDM_mixed':
        from pyHalo.PresetModels.wdm import WDM_mixed
        return WDM_mixed
    elif name == 'WDMGeneral':
        from pyHalo.PresetModels.wdm import WDMGeneral
        return WDMGeneral
    elif name == "DMGalacticus":
        from pyHalo.PresetModels.external import DMFromGalacticus
        return DMFromGalacticus
    else:
        raise Exception('preset model '+ str(name)+' not recognized!')






















from pyHalo.Rendering.subhalos import Subhalos
from pyHalo.Rendering.line_of_sight import LineOfSight, LineOfSightNoSheet
from pyHalo.Rendering.two_halo import TwoHaloContribution
import numpy as np

class HaloPopulation(object):

    """
    This class handles combinations of different preset halo populations. Models currently implemented are LINE_OF_SIGHT,
    SUBHALOS, AND TWO_HALO. See the documentation of each of these classes for details about what kinds of halos it
    adds to the lens model.
    """

    def __init__(self, model_list, mass_function_class_list, kwargs_mass_function_list,
                 spatial_distribution_class_list, kwargs_spatial_distribution,
                 lens_cosmo, geometry, lens_plane_redshift_list=None, redshift_spacings=None,
                 two_halo_Lazar_correction=True):

        """

        :param model_list: a list of names for different halo populations (e.g. ['LINE_OF_SIGHT', 'SUBHALOS', ...])
        :param mass_function_class_list: a list of non-instatiated mass function classes (see Rendering/MassFunctions)
        :param kwargs_mass_function_list: keyword arguments for the mass function classes
        :param spatial_distribution_class_list: a list of non-instantiated spatial distribution classes
        (see Rendering/SpatialDistributions)
        :param kwargs_spatial_distribution: keyword arguments for the spatial distribution classes
        :param lens_cosmo: an instance of LensCosmo (see Halos.lens_cosmo)
        :param geometry: an instance of Geometry class (see Cosmology.geometry)
        :param lens_plane_redshift_list: redshifts at which to render line-of-sight halos
        :param redshift_spacings: spacing between redshift planes
        :param two_halo_Lazar_correction: implements the correction derived by Lazar et al. 2021 for the two-halo term
        (see functions in the two_halo.py file)
        """

        self.rendering_classes = []
        for i, model_name in enumerate(model_list):

            mass_function_model_class = mass_function_class_list[i]
            kwargs_model = kwargs_mass_function_list[i]

            if model_name == 'SUBHALOS':
                if spatial_distribution_class_list[i].name in ['UNIFORM', 'LENS_CONE_UNIFORM']:
                    spatial_distribution_model = spatial_distribution_class_list[i](
                        **kwargs_spatial_distribution[i])
                else:
                    spatial_distribution_model = spatial_distribution_class_list[i].from_Mhost(**kwargs_spatial_distribution[i])
            else:
                spatial_distribution_model = spatial_distribution_class_list[i](**kwargs_spatial_distribution[i])

            if model_name == 'LINE_OF_SIGHT':
                model = LineOfSight(mass_function_model_class, kwargs_model, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshift_list, redshift_spacings)
            elif model_name == 'LINE_OF_SIGHT_NOSHEET':
                model = LineOfSightNoSheet(mass_function_model_class, kwargs_model, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshift_list, redshift_spacings)
            elif model_name == 'SUBHALOS':
                model = Subhalos(mass_function_model_class, kwargs_model, spatial_distribution_model,
                 geometry, lens_cosmo)
            elif model_name == 'TWO_HALO':
                model = TwoHaloContribution(mass_function_model_class, kwargs_model, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshift_list, redshift_spacings, two_halo_Lazar_correction)
            else:
                raise Exception('model '+str(model_name)+' not recognized. ')

            self.rendering_classes.append(model)

    def render(self):

        """
        Generates the masses, positions, 3D position inside host halo (if subhalo), and a list of bools that indicate
        whether each object is a subhalo

        :return: masses [M_sun], x coordinate [arcsec], y coordinate [arcsec], 3D position inside host halo [kpc],
        redshifts, list of bools indicating whether each object is a subhalo
        """
        masses = np.array([])
        x = np.array([])
        y = np.array([])
        r3d = np.array([])
        redshifts = np.array([])
        is_subhalo_flag = []

        for model in self.rendering_classes:

            m, _x, _y, _r3d, _z, sub_flag = model.render()
            masses = np.append(masses, m)
            x = np.append(x, _x)
            y = np.append(y, _y)
            r3d = np.append(r3d, _r3d)
            redshifts = np.append(redshifts, _z)
            is_subhalo_flag += sub_flag

        return masses, x, y, r3d, redshifts, is_subhalo_flag



from pyHalo.Halos.halo_base import Halo
import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import RegularGridInterpolator


def tnfw_mass_fraction(tau, c):
    """
    This function returns the fraction = final_mass/initial_mass, assuming a truncated NFW profile
    :param tau: the truncation radius in units of the scale radius
    :param c: the halo concentration
    """
    x = c
    Rs = 1.0
    r_trunc = tau * Rs
    func = (r_trunc ** 2 * (-2 * x * (1 + r_trunc ** 2) + 4 * (1 + x) * r_trunc * np.arctan(x / r_trunc) -
                            2 * (1 + x) * (-1 + r_trunc ** 2) * np.log(Rs) + 2 * (1 + x) * (-1 + r_trunc ** 2) * np.log(
            Rs * (1 + x)) +
                            2 * (1 + x) * (-1 + r_trunc ** 2) * np.log(Rs * r_trunc) -
                            (1 + x) * (-1 + r_trunc ** 2) * np.log(Rs ** 2 * (x ** 2 + r_trunc ** 2)))) / (
                   2. * (1 + x) * (1 + r_trunc ** 2) ** 2)
    mass_loss = func / (np.log(1 + c) - c / (1 + c))
    return mass_loss


def tau_mf_interpolator():
    N = 250
    tau = np.logspace(-3.5, 2.5, N)
    concentration = np.linspace(1.0, 200.0, N)

    log10_mass_fraction_1d = np.linspace(-4, -0.001, N)
    log10tau_2d = np.zeros((N, N))

    # This computes the value of tau that corresponds to each pair of (concentration, mass_loss)
    for i, con_i in enumerate(concentration):
        mfinal = tnfw_mass_fraction(tau, con_i)
        log10final_mass = np.log10(mfinal)
        mfinterp = interp1d(log10final_mass, np.log10(tau), fill_value='extrapolate')

        for j, log10_mass_j in enumerate(log10_mass_fraction_1d):
            log10tau_2d[i, j] = float(mfinterp(log10_mass_j))

    interp_points = (concentration, log10_mass_fraction_1d)
    interpolator = RegularGridInterpolator(interp_points, log10tau_2d, fill_value=None, bounds_error=False)
    return interpolator

_truncation_radius_interpolator = tau_mf_interpolator()

class TNFWSubhaloEmulator(Halo):
    """
    Defines a truncated NFW halo that is a subhalo of the host dark matter halo
    """

    def __init__(self, infall_mass, x, y, final_bound_mass, infall_concentration, redshift,
                 lens_cosmo_instance, unique_tag=None):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance

        r3d = None
        profile_definition = 'TNFW'
        sub_flag = True
        args = None
        if unique_tag is None:
            unique_tag = np.random.rand()

        # set the concentration
        self._c = infall_concentration
        self._bound_mass_fraction = final_bound_mass / infall_mass
        self._kpc_per_arcsec_at_z = self._lens_cosmo.cosmo.kpc_proper_per_asec(redshift)
        x_arcsec = x / self._kpc_per_arcsec_at_z
        y_arcsec = y / self._kpc_per_arcsec_at_z
        super(TNFWSubhaloEmulator, self).__init__(infall_mass, x_arcsec, y_arcsec, r3d, profile_definition, redshift,
                                                  sub_flag,
                                                  lens_cosmo_instance, args, unique_tag)

    @property
    def c(self):
        """
        :return: the halo concentration
        """
        return self._c

    @property
    def z_eval(self):
        """
        Returns the halo redshift
        """

        return self.z

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):
            [concentration, rt] = self.profile_args
            # evaluate density parameters at the time of lensing
            _rhos_mpc, _rs_mpc, _ = self._lens_cosmo.nfwParam_physical(self.mass, concentration, self.z_eval)
            # convert to angles at the time of lensing (deflector redshift)
            Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle(self.mass, concentration, self.z_eval)
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            theta_Rs = np.round(theta_Rs, 10)
            r_trunc_arcsec = rt / self._kpc_per_arcsec_at_z
            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                       'center_x': x, 'center_y': y, 'r_trunc': r_trunc_arcsec}]
            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            point = (self.c, np.log10(self._bound_mass_fraction))
            Rs_angle, _ = self._lens_cosmo.nfw_physical2angle(self.mass, self.c, self.z)
            log10_tau = float(_truncation_radius_interpolator(point))
            rt_over_rs = 10 ** log10_tau
            truncation_kpc = Rs_angle * rt_over_rs * self._kpc_per_arcsec_at_z
            self._profile_args = (self.c, truncation_kpc)

        return self._profile_args

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_params_physical'):
            [concentration, rt] = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            self._params_physical = {'rhos': rhos, 'rs': rs, 'r200': r200, 'r_trunc_kpc': rt}

        return self._params_physical

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['TNFW']



from pyHalo.Halos.halo_base import Halo
import numpy as np
from pyHalo.Halos.tnfw_halo_util import tnfw_mass_fraction
from scipy.integrate import quad

class TNFWFieldHalo(Halo):

    """
    The base class for a truncated NFW halo
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        mdef = 'TNFW'
        super(TNFWFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                           lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        if profile_args is None:
            c, rt = self.profile_args
        else:
            c, rt = profile_args
        rhos, rs, _ = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z_eval)
        tau = rt/rs
        x = r / rs
        rho_nfw = rhos / x / (1 + x) ** 2
        return rho_nfw * tau ** 2 / (tau ** 2 + x ** 2)

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['TNFW']

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        if not hasattr(self, '_params_physical'):
            [_, rt] = self.profile_args
            rhos, rs, r200 = self.nfw_params
            self._params_physical = {'rhos': rhos * self._rescale_norm, 'rs': rs, 'r200': r200, 'r_trunc_kpc': rt}

        return self._params_physical

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):

            [concentration, rt] = self.profile_args
            Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle(self.mass, concentration, self.z)

            x, y = np.round(self.x, 4), np.round(self.y, 4)

            Rs_angle = np.round(Rs_angle, 10)
            theta_Rs = np.round(theta_Rs, 10)
            r_trunc_arcsec = rt / self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                      'center_x': x, 'center_y': y, 'r_trunc': r_trunc_arcsec}]

            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        if not hasattr(self, '_profile_args'):
            truncation_radius_kpc = self._truncation_class.truncation_radius_halo(self)

            self._profile_args = (self.c, truncation_radius_kpc)
        return self._profile_args

class TNFWSubhalo(TNFWFieldHalo):
    """
    Defines a truncated NFW halo that is a subhalo of the host dark matter halo
    """

    @property
    def bound_mass(self):
        """
        Computes the mass inside the virial radius (with truncation effects included)
        :return: the mass inside r = c * r_s
        """
        if hasattr(self, '_kwargs_lenstronomy'):
            tau = self._kwargs_lenstronomy[0]['r_trunc'] / self._kwargs_lenstronomy[0]['Rs']
        else:
            params_physical = self.params_physical
            tau = params_physical['r_trunc_kpc'] / params_physical['rs']
        f = tnfw_mass_fraction(tau, self.c)
        return f * self.mass



import numpy as np
from lenstronomy.LensModel.Profiles.cnfw import CNFW
from pyHalo.Rendering.SpatialDistributions.base import SpatialDistributionBase
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.utilities import inverse_transform_sampling
from scipy.interpolate import interp1d


class ProjectedNFW(SpatialDistributionBase):
    name = 'PROJECTEDNFW'
    """
    This class approximates sampling from a full 3D NFW profile by
    sampling the projected mass of a cored NFW profile in 2D, and then sampling
    the z coordinate from an NFW profile. This method should approximate sampling from a cored NFW when the maximum
     2D radius is much less than the maximum 3D radius
    """

    def __init__(self, rmax2d_arcsec, rs_arcsec, r_core_arcsec, r_200_arcsec, arcsec_to_kpc):

        """

        :param Rs: the scale radius of the host dark matter halo [kpc]
        :param rmax2d: the maximum projected 2D radius where halos are rendered [arcsec]
        :param rmax3d: the virial radius of the host dark matter halo [kpc]
        :param r_core_parent: the core radius of the host dark matter halo [kpc]
        """

        self._rs_arcsec = rs_arcsec
        self._rmax2d = rmax2d_arcsec
        self._rmax3d = r_200_arcsec
        self._x3dmax = r_200_arcsec / rs_arcsec
        self._rcore_arcsec = r_core_arcsec
        self._cnfw_profile = CNFW()
        self._arcsec_to_kpc = arcsec_to_kpc
        super(ProjectedNFW, self).__init__()

    @classmethod
    def from_Mhost(cls, m_host, zlens, rmax2d_arcsec, r_core_units_rs, lens_cosmo, c_host=None):
        """

        :param m_host: host halo mass
        :param zlens: the deflector redshift
        :param rmax2d_arcsec: the maximum radius in projection to rendering subhalos
        :param r_core_units_rs: the core size in units rs of a cored nfw profile (subhalos are distributed
        following a cored nfw profile)
        :param lens_cosmo: an instance of LensCosmo
        :param c_host: host halo concentration
        :return:
        """
        if c_host is None:
            c_model = ConcentrationDiemerJoyce(lens_cosmo)
            c_host = c_model.nfw_concentration(m_host, zlens)
        _, rs_mpc, r_200_mpc = lens_cosmo.nfwParam_physical(m_host, c_host, zlens)
        arcsec_to_kpc = lens_cosmo.cosmo.kpc_proper_per_asec(zlens)
        rs_arcsec = rs_mpc * 1000 / arcsec_to_kpc
        r_200_arcsec = r_200_mpc * 1000 / arcsec_to_kpc
        r_core_arcsec = r_core_units_rs * rs_arcsec
        return ProjectedNFW(rmax2d_arcsec, rs_arcsec, r_core_arcsec, r_200_arcsec, arcsec_to_kpc)

    def draw(self, N, z_plane=None):

        if N == 0:
            return [], [], []

        x2d = np.linspace(1e-3 * self._rmax2d, self._rmax2d, 20000)
        args = (0.0, self._rs_arcsec, 1.0, self._rcore_arcsec)
        rho2d = self._cnfw_profile.density_2d(x2d, *args)
        rho2d_integral = [rho2d[i] * (x2d[i+1]**2 - x2d[i]**2) for i in range(0, len(x2d)-1)]
        function_2d = interp1d(x2d[0:-1], rho2d_integral)
        r2d_arcsec = inverse_transform_sampling(x2d[0:-1], function_2d, (), N)
        r2d_kpc = r2d_arcsec * self._arcsec_to_kpc
        theta = np.random.uniform(0, 2*np.pi, N)
        x_kpc, y_kpc = r2d_kpc * np.cos(theta), r2d_kpc * np.sin(theta)

        x3d = np.linspace(1e-3 * self._rmax2d, self._rmax3d, 20000)
        args = (self._rs_arcsec, 1.0, self._rcore_arcsec)
        rho3d = self._cnfw_profile.density(x3d, *args)
        rho3d_integral = [rho3d[i] * (x3d[i+1]**3 - x3d[i]**3) for i in range(0, len(x3d)-1)]
        function_3d = interp1d(x3d[0:-1], rho3d_integral)
        r3d_arcsec = inverse_transform_sampling(x3d[0:-1], function_3d, (), N)
        r3d_kpc = r3d_arcsec * self._arcsec_to_kpc

        return x_kpc, y_kpc, r3d_kpc

    # def draw(self, N, z_plane=None):
    #
    #     if N == 0:
    #         return [], [], []
    #
    #     x2d = np.linspace(1e-3 * self._rmax2d, self._rmax2d, 50000)
    #     function_2d = self._cnfw_profile.density_2d
    #     args = (0.0, self._rs_arcsec, 1.0, self._rcore_arcsec)
    #     r2d_arcsec = inverse_transform_sampling(x2d, function_2d, args, N)
    #     r2d_kpc = r2d_arcsec * self._arcsec_to_kpc
    #     theta = np.random.uniform(0, 2*np.pi, N)
    #     x_kpc, y_kpc = r2d_kpc * np.cos(theta), r2d_kpc * np.sin(theta)
    #
    #     x3d = np.linspace(1e-3 * self._rmax2d, self._rmax3d, 50000)
    #     function_3d = self._cnfw_profile.density
    #     args = (self._rs_arcsec, 1.0, self._rcore_arcsec)
    #     r3d_arcsec = inverse_transform_sampling(x3d, function_3d, args, N)
    #     r3d_kpc = r3d_arcsec * self._arcsec_to_kpc
    #
    #     return x_kpc, y_kpc, r3d_kpc




import numpy as np
from pyHalo.Rendering.SpatialDistributions.base import SpatialDistributionBase

class LensConeUniform(SpatialDistributionBase):
    name = 'LENS_CONE_UNIFORM'
    """
    This class generates samples drawn uniformly in two dimensions out to maximum radius
    r(z) = 0.5 * cone_opening_angle * f(z), where cone_opening_angle is the opening angle of the rendering volume
    specified when creating the realization, and f(z) is given by geometry.rendering_scale(z) and depends on the rendering
    geometry. For example, if the rendering volume is definied as DOUBLE_CONE (the default setting), then f(z) = 1
    for z < z_lens and decreases approaching the source redshift at z > z_lens. As the effect of the rendering geometry
    is built into this class, it is best used to generate halo positions distributed randomly in two dimensions
    along the line of sight.
    """

    def __init__(self, cone_opening_angle, geometry):

        """

        :param cone_opening_angle: the opening angle of the rendering volume [arcsec]

        If the rendering geometry is DOUBLE_CONE (default) then this is the opening angle of the cone

        If the rendering geometry is set to CYLINDER, then this is sets the comoving radius of the cylinder such that
        a ray that moves in a straight line from the observer to the lens plane at an angle cone_opening_angle inersects
        the edge of the cylinder at z_lens.

        :param geometry: an instance of Geometry (Cosmology.geometry)
        """
        self._cosmo_geometry = geometry
        self._uni = Uniform(0.5 * cone_opening_angle, geometry)
        super(LensConeUniform, self).__init__()

    @classmethod
    def from_Mhost(cls, *args, **kwargs):
        """

        :param rmax2d_arcsec:
        :param geometry:
        :return:
        """
        raise Exception('Spatial distribution class LensConeUniform not currently implemented for subhalos')

    def draw(self, N, z_plane, center_x=0, center_y=0):

        """
        Generates samples in two dimensions out to a maximum radius r = 0.5 * cone_opening_angle * f(z)
        where f(z) = geometry.rendering_scale(z)
        :param N: number of objects to generate
        :param z_plane: the redshift where the objects are being placed (used to compute the conversion to physical kpc)
        :param center_x: the x-center of the rendering area [arcsec]
        :param center_y: the y-center of the rendering area [arcsec]
        :return: the x and y coordinates sampled in 2D [kpc]
        """
        if N == 0:
            return np.array([]), np.array([])

        rescale = self._cosmo_geometry.rendering_scale(z_plane)

        x_kpc, y_kpc = self._uni.draw(N, z_plane, rescale=rescale,
                                        center_x=center_x, center_y=center_y)

        return x_kpc, y_kpc

class Uniform(SpatialDistributionBase):
    name = 'UNIFORM'
    """
    This class generates samples distributed uniformly in two dimensions out to a radius 0.5 * cone_opening_angle
    """

    def __init__(self, rmax2d_arcsec, geometry):

        """

        :param rmax2d_arcsec: the maximum radius to render objects [arcsec]
        :param geometry: an instance of Geometry (Cosmology.geometry)
        """
        self.rmax2d_arcsec = rmax2d_arcsec
        self._geo = geometry
        super(Uniform, self).__init__()

    @classmethod
    def from_Mhost(cls, *args, **kwargs):
        """

        :param rmax2d_arcsec:
        :param geometry:
        :return:
        """
        raise Exception('Spatial distribution class Uniform not currently implemented for subhalos')

    def draw(self, N, z_plane, rescale=1.0, center_x=0, center_y=0):

        """
        Generate samples distributed uniformly in two dimensions
        :param N: number of objects to generate
        :param z_plane: the redshift where the objects are being placed (used to compute the conversion to physical kpc)
        :param rescale: rescales the maximum rendering radius
        :param center_x: the x-center of the rendering area [arcsec]
        :param center_y: the y-center of the rendering area [arcsec]
        :return: the x and y coordinates sampled in 2D [kpc]
        """
        if N == 0:
            return [], []

        angle = np.random.uniform(0, 2 * np.pi, int(N))

        rmax = self.rmax2d_arcsec * rescale

        r = np.random.uniform(0, rmax ** 2, int(N))

        x_arcsec = r ** .5 * np.cos(angle)
        y_arcsec = r ** .5 * np.sin(angle)

        x_arcsec += center_x
        y_arcsec += center_y

        kpc_per_asec = self._geo.kpc_per_arcsec(z_plane)
        x_kpc, y_kpc = x_arcsec * kpc_per_asec, y_arcsec * kpc_per_asec

        return np.array(x_kpc), np.array(y_kpc)



import numpy as np
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tnfw_halo_util import tau_mf_interpolation
from colossus.lss import peaks
from colossus.halo import splashback
from pyHalo.Halos.galacticus_truncation.interp_mass_loss import InterpGalacticus, InterpGalacticusKeeley24
from pyHalo.Halos.galacticus_truncation.transfer_function_density_profile import compute_r_te_and_f_t


class TruncationSplashBack(object):

    def __init__(self, lens_cosmo):
        """
        This computes the splashback radius of a halo as the truncation radius (appropriate for field halos)
        See Diemer (2020)
        :param lens_cosmo: an instance of LensCosmo
        """
        self._lens_cosmo = lens_cosmo

    def truncation_radius_halo(self, halo):
        """
        Thiis method computes the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.z)

    def truncation_radius(self, halo_mass, z):
        """
        Computes the truncation radius as the splashback radius an NFW halo
        :param halo_mass: halo mass (m200 with respect to critical density at z)
        :param z: redshift
        :param halo_concentration: the halo concentration
        :param lens_cosmo: an instance of LensCosmo
        :return: the truncation radius
        """
        h = self._lens_cosmo.cosmo.h
        nu200m = peaks.peakHeight(halo_mass * h, z)
        RspR200m_units_rvir, _ = splashback.splashbackModel('RspR200m', nu200m=nu200m, z=z, rspdef='sp-apr-mn')
        r200m_mpc = self._lens_cosmo.rN_M(halo_mass * h, z, 200.0) / self._lens_cosmo.cosmo.astropy.Om0 ** (1./3)
        rt_kpc = RspR200m_units_rvir * r200m_mpc * 1e3
        return rt_kpc

class TruncationRN(object):

    def __init__(self, lens_cosmo, LOS_truncation_factor=50):
        """
        This implements a tidal truncation at r_N, where N is some overdensity with respect to the critical density of
        the Universe at z
        :param lens_cosmo: an instance of LensCosmo
        :param LOS_truncation_factor: the multiple of the overdensity threshold at which to truncte the halo, e.g. N=200
        would truncate at r200
        """
        self._lens_cosmo = lens_cosmo
        self._N = LOS_truncation_factor

    def truncation_radius_halo(self, halo):
        """
        Thiis method computes the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.z)

    def truncation_radius(self, halo_mass, z):
        """
        Computes the radius r_N of an NFW halo
        :param halo_mass: halo mass (m200 with respect to critical density at z)
        :param z: redshift
        :param lens_cosmo: an instance of LensCosmo
        :return: the truncation radius
        """
        h = self._lens_cosmo.cosmo.h
        rN_physical_mpc = self._lens_cosmo.rN_M(halo_mass * h, z, self._N)
        return rN_physical_mpc*1000

class TruncationRoche(object):

    def __init__(self, lens_cosmo=None, RocheNorm=1.4, m_power=1./3, RocheNu=2./3):
        """
        This implements a tidal truncation for subhalos of the form

        r_t = norm * (m / 10^7) ^ m_power * (r3d/50 kpc)^r3d_power [kpc]

        The default values were calibrated for a subhalo inside an isothermal host potential
        https://ui.adsabs.harvard.edu/abs/2016PhRvD..94d3505C/abstract

        :param norm: the overall scale
        :param m_power: exponent for the dependence on halo mass
        :param r3d_power: exponent for the dependence on 3D position inside host
        """
        self._norm = RocheNorm
        self._m_power = m_power
        self._r3d_power = RocheNu

    def truncation_radius_halo(self, halo):

        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.r3d)

    def truncation_radius(self, subhalo_mass, subhalo_r3d):

        """
        :param M: m200
        :param r3d: 3d radial position in the halo (physical kpc)
        :return: the truncation radius in physical kpc
        """
        m_units_7 = subhalo_mass / 10 ** 7
        radius_units_50 = subhalo_r3d / 50
        rtrunc_kpc = self._norm * m_units_7 ** self._m_power * radius_units_50 ** self._r3d_power
        return np.round(rtrunc_kpc, 3)

class AdiabaticTidesTruncation(object):
    """
    An example of the type of class we want to create and implement in pyHalo
    """

    def __init__(self, lens_cosmo, log_m_host, z_host, mass_loss_interp):
        """

        :param lens_cosmo: an instance of the LensCosmo class
        :param log_m_host: the host halo mass
        :param z_host: the redshift of the host halo
        :param mass_loss_interp: an instance of RegularGridInterpolator
        The interpolator should take as input a point
        (log10(concentration), log10(r_pericenter), c_host) and return the
        asymptotic final bound mass divided by the infall mass
        """
        self._mass_loss_interp = mass_loss_interp
        min_max_c = [1.0, 10 ** 2.7]
        min_max_rperi = [10 ** -2.0, 10 ** 0.2]
        self._lens_cosmo = lens_cosmo
        cmodel = ConcentrationDiemerJoyce(self._lens_cosmo.cosmo.astropy, scatter=False)
        c_host = cmodel.nfw_concentration(10**log_m_host, z_host)
        self._chost = c_host
        self._host_dynamical_time = self._lens_cosmo.halo_dynamical_time(10**log_m_host, z_host, c_host)
        self._min_c = min_max_c[0]
        self._max_c = min_max_c[1]
        self._min_rperi = min_max_rperi[0]
        self._max_rperi = min_max_rperi[1]
        self._tau_mf_interpolation = tau_mf_interpolation()

    def truncation_radius_halo(self, halo):
        """
        This function solves for the truncation radius divided by the halo scale radius (tau) given an instance of
        Halo (see pyhalo.HaloModels.TNFW) and other arguments for the interpolation function

        :param halo: an instance of the Halo class (should be TNFW)

        :return tau: the truncation radius divided by the halo's scale radius
        """

        # compute the halo parameeters
        c = halo.c
        # now make sure that the points are inside the region where we computed the interpolation
        r_pericenter_over_r200 = np.absolute(halo.rperi_units_r200)
        r_pericenter_over_r200 = max(self._min_rperi, r_pericenter_over_r200)
        r_pericenter_over_r200 = min(self._max_rperi, r_pericenter_over_r200)
        c = max(self._min_c, c)
        c = min(self._max_c, c)
        point = (np.log10(c), np.log10(r_pericenter_over_r200), self._chost)
        # evaluate the mass loss
        log10mass_loss_fraction_asymptotic = float(self._mass_loss_interp(point))

        time_since_infall = halo.time_since_infall
        n_orbits = time_since_infall / self._host_dynamical_time
        mass_loss_fraction = self._temporal_mass_loss(10 ** log10mass_loss_fraction_asymptotic, n_orbits)
        log10mass_loss_fraction = np.log10(mass_loss_fraction)

        # solve for tau
        log10c = np.log10(c)
        point = (log10c, log10mass_loss_fraction)
        point = self._make_params_in_bounds_tau_evaluate(point)
        log10tau = float(self._tau_mf_interpolation(point))
        tau = 10 ** log10tau
        _, rs, _ = self._lens_cosmo.NFW_params_physical(halo.mass, halo.c, halo.z_eval)
        return tau * rs

    @staticmethod
    def _temporal_mass_loss(mass_loss_asymptotic, n_orbits):
        """
        This routine interpolates between zero mass loss at infall and the asymptotic value
        predicted by the adiabatic tides model
        """
        mass_loss = mass_loss_asymptotic * (1 / mass_loss_asymptotic) ** (1. / (1. + 0.6 * n_orbits))
        return mass_loss

    def _make_params_in_bounds_tau_evaluate(self, point):
        """
        This routine makes sure the arguments for the initerpolation are inside the domain of the function.
        """
        (log10c, log10mass_loss_fraction) = point
        log10c = max(np.log10(self._min_c), log10c)
        log10c = min(np.log10(self._max_c), log10c)
        log10mass_loss_fraction = max(-1.5, log10mass_loss_fraction)
        log10mass_loss_fraction = min(-0.01, log10mass_loss_fraction)
        return (log10c, log10mass_loss_fraction)

class TruncateMeanDensity(object):

    def __init__(self, lens_cosmo, median_rt_over_rs=2.0, c_power=3.0, intrinsic_scatter=0.0):
        """
        Implements a truncation model where the truncation radius depends on the infall concentration and orbital pericenter
        r_t / r_s = norm * (c / c_med)^c_power * (r_peri / (0.5 * r_200))^rperi_power
        :param lens_cosmo: an instance of LensCosmo
        :param median_rt_over_rs: the value of rt / rs for a halo that falls into the host with the median concetration
        for halos of its mass with an orbital pericenter of half the host's virial radius
        :param c_power: scales the dependence on infall concentration
        :param intrinsic_scatter: adds Gaussian scatter to rt_over_rs in unis of rt_over_rs, i.e.
        if intrinsic_scatter = s the scatter is given by Gaussian(mean = 0.0, sigma= s * rt_over_rs)
        """

        self._norm = median_rt_over_rs
        self._cpower = c_power
        self.lens_cosmo = lens_cosmo
        self._scatter = intrinsic_scatter
        self._concentration_cdm = ConcentrationDiemerJoyce(lens_cosmo.cosmo,
                                                           scatter=False)

    def truncation_radius_halo(self, halo):

        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        c_median = self._concentration_cdm.nfw_concentration(halo.mass, halo.z_eval)
        c_actual = halo.c
        return self.truncation_radius(halo.mass, halo.z, c_median, c_actual, self._scatter)

    def truncation_radius(self, halo_mass, halo_redshift, c_median, c_actual, intrinsic_scatter):
        """
        Computes the truncation radius from halo mass, halo redshift, median concentration at infall for halos of the
        given mass, and the actual concentration of the halo at infall, plus Gaussian scatter
        :param halo_mass: halo mass at infall
        :param halo_redshift: redshift of infall
        :param c_median: median concentration of halos of mass halo_mass at z = z_infall in CDM
        :param c_actual: actual concentration of the halo at infall
        :param intrinsic_scatter: Gaussian scatter
        :return:
        """
        _rt_over_rs = self._norm * (c_actual / c_median) ** self._cpower
        if intrinsic_scatter > 0:
            rt_over_rs = abs(np.random.normal(_rt_over_rs, intrinsic_scatter * _rt_over_rs))
        else:
            rt_over_rs = _rt_over_rs
        _, rs, _ = self.lens_cosmo.NFW_params_physical(halo_mass, c_actual, halo_redshift)
        return rs * rt_over_rs

class ConstantTruncationArcsec(object):
    """A constant truncation radius in arcsec"""
    def __init__(self, lens_cosmo, rt_arcsec):
        self.rt_arcsec = rt_arcsec
        pass
    def truncation_radius_halo(self, halo):
        return self.rt_arcsec
    def truncation_radius(self, *args, **kwargs):
        return self.rt_arcsec

class TruncationGalacticusKeeley24(object):

    def __init__(self, lens_cosmo, c_host):
        """

        :param lens_cosmo:
        """

        self._chost = c_host
        self._lens_cosmo = lens_cosmo
        self._mass_loss_interp = InterpGalacticusKeeley24()
        self._tau_mf_interpolation = tau_mf_interpolation()

    @staticmethod
    def _make_params_in_bounds_tau_evaluate(point):
        """
        This routine makes sure the arguments for the interpolation are inside the domain of the function.
        """
        min_c, max_c = np.log10(1.0), np.log10(10 ** 2.7)
        (log10c, log10mass_loss_fraction) = point
        log10c = max(min_c, log10c)
        log10c = min(max_c, log10c)
        log10mass_loss_fraction = max(-8.0, log10mass_loss_fraction)
        log10mass_loss_fraction = min(-0.001, log10mass_loss_fraction)
        return (log10c, log10mass_loss_fraction)

    def truncation_radius_halo(self, halo):

        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """

        return self.truncation_radius(halo.mass, halo.c,
                                      halo.time_since_infall, self._chost, halo.z)

    def truncation_radius(self, halo_mass, infall_concentration,
                          time_since_infall, chost, z_eval_angles):
        """

        :param halo_mass:
        :param infall_concentration:
        :param time_since_infall:
        :param chost:
        :param z_eval_angles:
        :return:
        """
        log10c = np.log10(infall_concentration)
        log10mbound_over_minfall = self._mass_loss_interp(log10c, time_since_infall, chost)
        point = self._make_params_in_bounds_tau_evaluate((log10c, log10mbound_over_minfall))
        log10tau = float(self._tau_mf_interpolation(point))
        tau = 10 ** log10tau
        _, rs, _ = self._lens_cosmo.NFW_params_physical(halo_mass,
                                                        infall_concentration,
                                                        z_eval_angles)
        return tau * rs

class TruncationGalacticus(object):

    def __init__(self, lens_cosmo, c_host):
        """

        :param lens_cosmo: an instance of LensCosmo
        :param c_host: host halo concentration at z=0.5
        """
        self._chost = c_host
        self._lens_cosmo = lens_cosmo
        self._mass_loss_interp = InterpGalacticus()
        self._tau_mf_interpolation = tau_mf_interpolation()

    @staticmethod
    def _make_params_in_bounds_tau_evaluate(point):
        """
        This routine makes sure the arguments for the interpolation are inside the domain of the function.
        """
        min_c, max_c = np.log10(1.0), np.log10(10 ** 2.7)
        (log10c, log10mass_loss_fraction) = point
        log10c = max(min_c, log10c)
        log10c = min(max_c, log10c)
        log10mass_loss_fraction = max(-8.0, log10mass_loss_fraction)
        log10mass_loss_fraction = min(-0.001, log10mass_loss_fraction)
        return (log10c, log10mass_loss_fraction)

    def truncation_radius_halo(self, halo):
        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """

        halo_mass = halo.mass
        infall_concentration = halo.c
        time_since_infall = halo.time_since_infall
        log10c = np.log10(infall_concentration)
        log10mbound_over_minfall = self._mass_loss_interp(log10c, time_since_infall, self._chost)
        m_bound = halo_mass * 10 ** log10mbound_over_minfall
        _, rs, r200 = self._lens_cosmo.NFW_params_physical(halo_mass,
                                                           infall_concentration,
                                                           halo.z)
        r_te, f_t = compute_r_te_and_f_t(m_bound, halo_mass, r200, infall_concentration)
        halo.rescale_normalization(f_t)
        return r_te

    def truncation_radius(self, halo_mass, infall_concentration,
                          time_since_infall, z_eval_angles):
        """

        :param halo_mass:
        :param infall_concentration:
        :param time_since_infall:
        :param chost:
        :param z_eval_angles:
        :return:
        """
        log10c = np.log10(infall_concentration)
        log10mbound_over_minfall = self._mass_loss_interp(log10c, time_since_infall, self._chost)
        m_bound = halo_mass * 10 ** log10mbound_over_minfall
        _, rs, r200 = self._lens_cosmo.NFW_params_physical(halo_mass,
                                                        infall_concentration,
                                                        z_eval_angles)
        r_te, _ = compute_r_te_and_f_t(m_bound, halo_mass, r200, infall_concentration)
        return r_te



